/*************************************************
 * CAMELOID Secure End-to-End Telephony System
 * Copyright (C) 2005  Dr. Tek                     
 ************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <sys/ioctl.h>

#ifdef	 LINUX
#include <endian.h>
#include <linux/soundcard.h>
#endif

#ifdef	 BSD
#include <sys/endian.h>
#include <soundcard.h>
#endif

#include <ctype.h>
#include <sys/mman.h>
#include <termios.h>


#include "libs/sha256.h"
#include "libs/ecrypt-sync.h"
#include "libs/phelix.h"
#include "libs/tiger.h"
#include "libs/md5.h"
#include "libs/py.h"
#include "libs/serpent.h"
#include "libs/twofish.h"
#include "libs/cast6.h"
#include "libs/aes.h"
#include <pthread.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <getopt.h>
#include "cameloid_algo.h"

#ifdef	VIDEO_CAP
#include <linux/videodev.h>
#include <jpeglib.h>
#endif

#ifndef		RANDOM_DEVICE
#define		RANDOM_DEVICE		"/dev/random"
#endif

#ifndef		URANDOM_DEVICE
#define		URANDOM_DEVICE		"/dev/urandom"
#endif


#if		(!VIDEO_CAP && !IM_CAP)
#ifndef		SOUND_CARD_DEVICE	
#define		SOUND_CARD_DEVICE	"/dev/dsp"
#endif
#define		CAMELOID_SOUND_DEVICE		SOUND_CARD_DEVICE
#ifndef         RING_SOUND_FILE           
#define         RING_SOUND_FILE         "./ring.wav"
#endif
#define         SOUND_SPEED             8000
#define         SOUND_BITS              16
#define         SOUND_MONO              1
#endif



#ifdef		VIDEO_CAP
#ifndef		VIDEO_DEVICE
#define		VIDEO_DEVICE		"/dev/video0"
#endif
#define		CAMELOID_VIDEO_DEVICE		VIDEO_DEVICE
#endif

#ifdef 		DEFINE_VAR_ALIASES
typedef unsigned char  u8;
typedef unsigned short u16;
typedef unsigned long  u32;
#endif


#define         RD_BUFFER_SIZE          	1024
#define		CAMELOID_MAX_USERNAME		10

/* Prototypes */

__inline__ void cameloid_destroy_heap(u8 *, u32);
__inline__ int  cameloid_open_random(void);
__inline__ int  cameloid_open_urandom(void);
__inline__ void cameloid_get_r_random_bytes(int, u8 *, u32);
__inline__ void cameloid_get_p_random_bytes(int, u8 *, u32);
__inline__ u8 * cameloid_generate_random_key(int, u32);
__inline__ void cameloid_destroy_memory(u8 *, u32);
__inline__ void cameloid_print_error(u8 *);
__inline__ int  cameloid_open_sound(u8 *, int);
__inline__ void cameloid_read_voice(int,u8 *,u32);
__inline__ void cameloid_write_voice(int,u8 *,u32);
__inline__ u8 * cameloid_hash_and_show(const u8 * , u32);
__inline__ void cameloid_generate_hash(const u8 *, u32, u8 *);
__inline__ int  cameloid_lock_all_memory(void);
__inline__ u8 * cameloid_get_memory(u32);
__inline__ void cameloid_free_memory(u8 *, u32);
__inline__ int  cameloid_check_sound_card_lock(void);
__inline__ int  cameloid_lock_sound_card(void);
__inline__ void do_system_beep(void);
__inline__ void cameloid_unlock_sound_card(void);


#ifdef	VIDEO_CAP

#define         COMPRESS_BUF_SIZE               (32*1024)

__inline__ int  cameloid_open_video(u8 *);
__inline__ int  cameloid_read_video(int,u32 *, u8 *, void *);
__inline__ int  cameloid_decompress_mjpeg(u8 *, u8 *, u32);

#endif



#ifdef	IM_CAP
#define	 	IM_BUF_SIZE			(256 + 15)
#endif
	

/* Some of our globals */
int 	active_socket,verbose,sound_switch,data_port,
	use_strong_rand,send_delay,card_duplex,choice_algo,my_lock;
u8  	*NK, *SK;


#ifdef		VIDEO_CAP
#define		VERSION		"0.0.4b\0"
#endif

#ifdef		IM_CAP
#define		VERSION		"0.0.3a\0"
#endif

#if		(!VIDEO_CAP && !IM_CAP)
#define         VERSION         "0.1.8b\0"
#endif


