/*************************************************
 * CAMELOID Secure End-to-End Telephony System
 * Copyright (C) 2005  Dr. Tek                     
 ************************************************/

#include "cameloid.h"
#include "cameloid_protocol.h"
#include "cameloid_memory.h"


/* 
 *
 *	
 *	Memory Sanitizing Functions
 *
 *
 */

__inline__
void cameloid_destroy_heap(u8 *data, u32 size)
{
        if(data)
        {
                memset(data,0,size);
                free(data);
        }
}


__inline__
void cameloid_destroy_memory(u8 *data, u32 size)
{
        if(data)
        {
		memset(data,0,size);
	}	
}

__inline__
int cameloid_lock_all_memory(void)
{

	if((mlockall(MCL_CURRENT|MCL_FUTURE))<0)
	{
		printf("Unable to secure memory\n");
		return -1;
			
	}

	return 0;	
}


__inline__
u8 * cameloid_get_memory(u32 size)
{
	u8 *p = NULL;
	
	p = (u8 *) calloc(1,size);
	if(!p)
	{
		printf("Cannot allocate %lu bytes of memory\n",size);
		return NULL;
	}

	return (u8 *) p;

}


__inline__
void cameloid_free_memory(u8 *ptr, u32 size)
{
	cameloid_destroy_heap(ptr,size);
}	


__inline__
int cameloid_allocate_memory_struct(cameloid_mstruct_t *m)
{

	if((m->NK = cameloid_get_memory(CAMELOID_MAX_NK)) == NULL)
		return -1;
	
	if((m->NK_TEMP = cameloid_get_memory(CAMELOID_MAX_NK)) == NULL)
        {
		free(m->NK);
	        return -1;
	}

	if((m->SK = cameloid_get_memory(CAMELOID_MAX_SK)) == NULL)
	{
		free(m->NK_TEMP); 
		free(m->NK);
		return -1;
	}

	return 0;
}


__inline__
void cameloid_destroy_memory_struct(cameloid_mstruct_t *m)
{
	
	if(m->NK)
	{
		cameloid_destroy_memory(m->NK,CAMELOID_MAX_NK);
		m->NK = NULL;
	}

	if(m->NK_TEMP)
	{
		cameloid_destroy_memory(m->NK_TEMP,CAMELOID_MAX_NK);
		m->NK_TEMP = NULL;
	}

	if(m->SK)
	{
		cameloid_destroy_memory(m->SK,CAMELOID_MAX_SK);
		m->SK = NULL;
	}

}




/* Memory Allocation */

void *
cameloid_alloc(u32 size)
{
	return cameloid_get_memory(size);

}
 
