/*************************************************
 * CAMELOID Secure End-to-End Telephony System                   
 * Copyright (C) 2005  Dr. Tek
 ************************************************/

#include "cameloid.h"
#include "cameloid_memory.h"
#include "cameloid_voice.h"
#include "cameloid_protocol.h"
#include "cameloid_functions.h"


/* Globals */
int 			 randomfd, urandomfd, s, target_port;
 
u8 			*sc_read_dev  = CAMELOID_SOUND_DEVICE,  
   			*sc_write_dev = CAMELOID_SOUND_DEVICE;

cameloid_kc_t  		*keystruct;
cameloid_mstruct_t 	 memctx;

__inline__
void kill_connection(int sig)
{
	static int terminprogress = 0;
	struct termios termset;

	/* This is done to ensure that the terminal will 
	   have the echo flags restored in case of a pre-mature
	   cancellation */

	 tcgetattr(fileno(stdin),&termset);
	 termset.c_lflag |= (ECHO | ECHOE | ECHOK | ECHONL);
	 tcsetattr(fileno(stdin),TCSANOW,&termset);	
        
	if(!terminprogress)
	{
		terminprogress = 1;
	
		printf("\nTerminating connection\n\n");

		if(my_lock)
		{
			cameloid_unlock_sound_card();
		}

		cameloid_destroy_memory_struct(&memctx);
		if(keystruct)
		{
			cameloid_destroy_memory((u8 *) keystruct, sizeof(cameloid_kc_t));
		}

		close(s);
		exit(0);
	}

}

__inline__
int cameloid_open_conn(u32 addr, u8 *username)
{
	int 	 	 	 sklen = sizeof(struct sockaddr_in),  bufsize = 40000,
			 	 s  = 0, data_socket = 0, cprecv = 0, nduplex = 0, 
				 soundfd = 0, out_socket = 0;

	struct sockaddr_in 	 sk, sk2, sk3;
	u8 			 *buf, *nonce,*digest, 
				 *ptr = NULL,   hash[32];

	cameloid_init_pkt_t 	 init_h;
	cameloid_sk_xchg_t 	*sk_h, sk_w_h;
	cameloid_cr_pkt_t 	*cr, snd_cr_h;
	cameloid_auth_t 	*ath_w_p, *ath_r_p;
	cameloid_kc_t 		 ctx;
	fd_snd_t 		 fdsndst;
	pthread_t 		 pthread_st;		
	struct timeval	 	 tv;


	buf	= (u8 *) cameloid_alloc(RD_BUFFER_SIZE + 1);
	if(!buf)
	{
		return 0;
	}

	nonce	= (u8 *) cameloid_alloc(CAMELOID_NONCE_SIZE + 1);
	if(!nonce)
	{
		return 0;
	}

	digest	= (u8 *) cameloid_alloc(16); 
	if(!digest)
	{
		return 0;
	}


	cameloid_key_setup(choice_algo,NK,strlen(NK),&ctx);
	cameloid_destroy_memory(NK,strlen(NK));
	cameloid_destroy_memory((u8 *)&sk,  sklen);
	cameloid_destroy_memory((u8 *)&sk2, sklen);
	cameloid_destroy_memory((u8 *)&sk3, sklen);
	cameloid_destroy_memory((u8 *)&init_h, sizeof(init_h));

	

	keystruct = &ctx;

	if((data_socket = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP))<0)
        {                                                
                cameloid_print_error("Cannot open UDP network socket.");
                return -1;
        }

	
	if((out_socket  = dup(data_socket))<0)
        {       
                cameloid_print_error("Cannot open UDP network socket.");
                return -1;
        }

	tv.tv_sec        = CAMELOID_NET_UDP_TIMEOUT;
        tv.tv_usec       = 0; 


	setsockopt(data_socket, SOL_SOCKET,SO_RCVTIMEO,&tv,sizeof(struct timeval));                                                 
	setsockopt(out_socket,  SOL_SOCKET,SO_SNDBUF,  &bufsize, sizeof(int));
        setsockopt(data_socket, SOL_SOCKET,SO_RCVBUF,  &bufsize, sizeof(int));

        sk2.sin_port             = HTONS(data_port);
        sk2.sin_addr.s_addr      = addr;
        sk2.sin_family           = AF_INET;
	

	if((s = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP))<0)
	{
		cameloid_print_error("Error opening TCP socket.");
		return -1;

	}

	tv.tv_sec        = CAMELOID_NET_TIMEOUT;
        tv.tv_usec       = 0;

	setsockopt(s,SOL_SOCKET,SO_RCVTIMEO,&tv,sizeof(struct timeval));

	sk.sin_family 		= AF_INET;
	sk.sin_addr.s_addr 	= addr;
	sk.sin_port		= HTONS(target_port);

	clear_reset_screen();
	printf("Making secure connection to %s (Algorithm: %s)\n",inet_ntoa(sk.sin_addr),cameloid_algo_to_name(choice_algo));

	ath_w_p = (cameloid_auth_t *) &init_h.ath;
	
	/*******************************************************
	 ****  Fill in correct authentication information ******
	 *******************************************************/

	cameloid_get_p_random_bytes(urandomfd,nonce,CAMELOID_NONCE_SIZE);

	cameloid_build_auth_packet(ath_w_p,
				   nonce,
				    CAMELOID_PKT_INIT);

	ath_w_p->duplex 	= card_duplex;
	init_h.alg 		= choice_algo;
	
	memcpy(init_h.username,username,CAMELOID_MAX_USERNAME);

	/* obfuscate the username */
	cameloid_add_gf256(init_h.username,nonce,CAMELOID_MAX_USERNAME);

	if((connect(s,(struct sockaddr *) &sk,sizeof(struct sockaddr_in)))<0)
	{
		printf("Host is not accepting calls at this time\n");
		kill_connection(0);
		return 0;
	}

	if((connect(data_socket,(struct sockaddr *) &sk2, sizeof(struct sockaddr_in)))<0)
	{
		printf("Host is not accepting calls at this time\n");
		kill_connection(0);
		return 0;
	}

	getsockname(data_socket,(struct sockaddr *) &sk3, &sklen);
		
	init_h.data_port = sk3.sin_port;
	
	if((send(s,&init_h,PKT_IN_SIZE,0)) != PKT_IN_SIZE)
	{
		cameloid_print_error("Error sending initial packet.");
		kill_connection(0);
		return 0;
	}
	
	if(verbose)
	{
		printf("Awaiting challenge packet response\n");
		fflush(stdout);
	}

	if((recv(s,buf, PKT_CR_SIZE,0)) != PKT_CR_SIZE)
	{
		printf("Host did not respond in expected time\n");
		kill_connection(0);
		return 0;		
	}
	else
	{
		cr = (cameloid_cr_pkt_t *) buf;
		
		ath_r_p  	= (cameloid_auth_t *) (&cr->ath);
	
		if(cr->ath.pkt_type == CAMELOID_PKT_CHALLENGE)
		{
			if(verbose)
			{
				printf("Challenge packet received sending response\n");				
				fflush(stdout);
			}

			cprecv = 1;
			nduplex = cr->ath.duplex;
			
			/* Verify, decrypt challenge and send back encrypted hash */										
			ptr = cr->data;

			cameloid_hmac_md5(ptr,
					  32,
					  ath_r_p->nonce,
					  CAMELOID_NONCE_SIZE,
					  ctx.keydatahash,
					  CAMELOID_MAC_KEYSIZE,
					  digest);

			if(memcmp(ath_r_p->hmac,digest,16))
			{
				/* This may be too paranoid of an assumption, Hence the word "Possible" */
				#ifdef SUPER_PARANOID
				cameloid_print_error("*Possible* injection attack, Invalid MAC computed for C/R packet."); 
				#endif
				cameloid_print_error("Response not valid, possibly due to incorrect NK specified.");
				close(s);
				return 0;
			}
				
			cameloid_crypt_data(CAMELOID_MODE_DECRYPT,
					    ptr,
					    32,
					    ath_r_p->nonce,&ctx);

			cameloid_generate_hash(ptr,32,hash);
		
			cameloid_get_p_random_bytes(urandomfd,nonce,CAMELOID_NONCE_SIZE);
		
			cameloid_crypt_data(CAMELOID_MODE_ENCRYPT,
					    hash,
					    32,
					    nonce,&ctx);
		
			cameloid_build_cr_packet(&snd_cr_h,hash,nonce,CAMELOID_PKT_RESPONSE);
		
			cameloid_hmac_md5(hash,
					  32,
					  nonce,
					  CAMELOID_NONCE_SIZE,
					  ctx.keydatahash,
					  CAMELOID_MAC_KEYSIZE,snd_cr_h.ath.hmac);
		
			cameloid_destroy_memory(hash,32);

			if((send(s,&snd_cr_h,PKT_CR_SIZE,0)) != PKT_CR_SIZE)
			{
				cameloid_print_error("Error sending Challenge/Response Packet.");
				kill_connection(0);
				return 0;
			}

			if((cameloid_net_read(s,buf,PKT_SK_SIZE)) != PKT_SK_SIZE)
			{
				printf("Call disconnected\n"); fflush(stdout);
				close(s);
				return 0;
			}

			
			sk_h = (cameloid_sk_xchg_t *) buf;
			ath_r_p = &sk_h->ath;
			
			if(ath_r_p->pkt_type != CAMELOID_PKT_SK)
			{
				printf("Session key segment packet not received\n"); fflush(stdout);
				close(s);	
				return 0;
			}
			else
			{	
				
				/* Verify SK packet, store it and generate our SK */
				if(verbose)
				{
					printf("Session key segment received\n");
					fflush(stdout);
				}
			
				cameloid_hmac_md5(sk_h->sk_segment,
						  32,
						  ath_r_p->nonce,
						  CAMELOID_NONCE_SIZE,
                                          	  ctx.keydatahash,
						  CAMELOID_MAC_KEYSIZE,
						  digest);

	                        if(memcmp(ath_r_p->hmac,digest,16))
        	                {
                	                printf("Possible injection attack, Invalid MAC computed for C/R packet.\n"); fflush(stdout);
                       		        kill_connection(0);
					return 0;
                        	}       
                    
				cameloid_crypt_data(CAMELOID_MODE_DECRYPT,
						    sk_h->sk_segment,
						    CAMELOID_MAX_SK_SEG,
						    ath_r_p->nonce,&ctx);
				
				if(verbose)
				{
					printf("Received SK segment hash -> %32s\n",cameloid_hash_and_show_md5(sk_h->sk_segment,CAMELOID_MAX_SK_SEG));
					fflush(stdout);
				}

				memcpy(SK,sk_h->sk_segment,CAMELOID_MAX_SK_SEG);

				/* Destroy generated session key segment */
				cameloid_destroy_memory(sk_h->sk_segment,CAMELOID_MAX_SK_SEG);				
				
				if(verbose)
				{
					printf("Generating %3d-bit random SK segment\n",CAMELOID_MAX_SK_SEG*8);
				}

				cameloid_get_r_random_bytes(randomfd,SK+CAMELOID_MAX_SK_SEG,CAMELOID_MAX_SK_SEG);
				cameloid_get_p_random_bytes(urandomfd,nonce,CAMELOID_NONCE_SIZE);
				cameloid_build_auth_packet(&sk_w_h.ath,nonce,CAMELOID_PKT_SK);
				memcpy(&sk_w_h.sk_segment,SK+CAMELOID_MAX_SK_SEG,CAMELOID_MAX_SK_SEG);
				if(verbose)
				{
					printf("Generated SK segment hash -> %32s\n",
						cameloid_hash_and_show_md5(SK+CAMELOID_MAX_SK_SEG,CAMELOID_MAX_SK_SEG));
				}

				cameloid_crypt_data(CAMELOID_MODE_ENCRYPT,
						    sk_w_h.sk_segment,
						    CAMELOID_MAX_SK_SEG,
						    nonce,&ctx);

				cameloid_hmac_md5(sk_w_h.sk_segment,
						  32,
						  nonce,
						  CAMELOID_NONCE_SIZE,
						  ctx.keydatahash,
                                                  CAMELOID_MAC_KEYSIZE,
						  sk_w_h.ath.hmac);

				cameloid_destroy_memory(nonce,CAMELOID_MAX_SK_SEG);

				/* Destroy old keychain  */
				cameloid_destroy_memory((u8 *) &ctx,sizeof(cameloid_kc_t));
				
				if(verbose)
				{
					printf("Exchanging session key segment\n");
					printf("Session key hash -> %32s\n",cameloid_hash_and_show_md5(SK,CAMELOID_MAX_SK));
					fflush(stdout);
				}
				
				if((send(s,&sk_w_h,PKT_SK_SIZE,0)) != PKT_SK_SIZE)
				{
					cameloid_print_error("Error exchange session key segment.");
					kill_connection(0);
					return 0;
				}
				
				close(s);
				active_socket = 255;
				
				/* Run newly created SK through the selected KSA */
				cameloid_key_setup(choice_algo,SK,CAMELOID_MAX_SK,&ctx);
				
				/* Destroy SK buf */
				cameloid_destroy_memory(SK,CAMELOID_MAX_SK);
				
				/* Secure voice session begins */
				if(nduplex == SOUND_CARD_HALF)
				{
					printf("Communicating securely in HALF duplex mode\n"); fflush(stdout);

					while(active_socket)
					{
		
						printf("Secure Talking Mode          \r"); fflush(stdout);			
						if((soundfd = cameloid_open_sound(sc_read_dev,O_RDONLY))<0)
						{
							kill_connection(0);
							return 0;
						}
					
						if((cameloid_read_half_duplex(data_socket,soundfd,urandomfd,sk2.sin_port,addr,&ctx))<0)
						{
							kill_connection(0);
							cameloid_close_sound(soundfd);
							return 0;
						}
						cameloid_close_sound(soundfd);

						if((soundfd = cameloid_open_sound(sc_write_dev,O_WRONLY))<0)
						{
							kill_connection(0);
							return 0;
						}
					
						printf("Secure Listening Mode         \r"); fflush(stdout);
						if((cameloid_write_half_duplex(data_socket,soundfd,urandomfd,&ctx))<0)
						{
							kill_connection(0);
							close(soundfd);
							return 0;
						}
						cameloid_close_sound(soundfd);

					}
				}
				else
				{
					fdsndst.writesocket  	= out_socket;
					fdsndst.readsocket	= data_socket;
					fdsndst.urandomfd 	= urandomfd;
					fdsndst.ctx 	  	= &ctx;
					fdsndst.sender_port	= sk2.sin_port;
					fdsndst.sender_addr	= addr;
					
					printf("Communicating securely in FULL duplex mode\n"); fflush(stdout);
					if((soundfd = cameloid_open_sound(SOUND_CARD_DEVICE,O_RDWR))<0)
                                         {
                                                kill_connection(0);
                                                 return 0;
                                         }
					 				
					fdsndst.soundfd  = soundfd;
				
					pthread_create(&pthread_st,NULL,(void *) cameloid_read_full_duplex,(void *) &fdsndst);
					cameloid_write_full_duplex((void *)&fdsndst);
					kill_connection(0);
					return 0;

				}

				}
				
			}
			
	}

	
	close(s);	
	cameloid_destroy_memory(buf,RD_BUFFER_SIZE + 1);
	return 0;
	

}

int main(int argc, char **argv)
{
	int c, i;
	u32 addr;
	struct hostent *target_host = NULL;
	u8 *hp = NULL;
	static u8 *username = NULL;

	close(2);	/* Close stderr because Speex is noisy */
	c = s = addr = use_strong_rand = choice_algo = 0;
	randomfd = urandomfd = card_duplex = verbose = my_lock = target_port = data_port = 0;
	send_delay = 1;

	printf("\n\tCAMELOID v%s\n\tSecure Network Telephony Client\n\n",VERSION);
	
	if(argc<2)
	{
		printf("Usage: %s -[uphrwvdsD] hostname\n",argv[0]);
		printf("Options:\n");
		printf("\t-p <port>\tAlternate cameloid server port (default port %5d)\n",CAMELOID_LISTEN_PORT);
		printf("\t-d <port>\tAlternate cameloid data port   (default port %5d)\n",CAMELOID_DATA_PORT);  
		printf("\t-u <username>\tUsername of sender\n");
		printf("\t-r <device>\tSound device for reading (default: %s)\n",sc_read_dev);
		printf("\t-w <device>\tSound device for writing (default: %s)\n",sc_write_dev);
		printf("\t-h\t\tForce half-duplex communication\n");
		printf("\t-v\t\tVerbose Mode\n");
		printf("\t-s\t\tUse strong random numbers (rijndael encrypted)\n");
		printf("\t-D <ms>\t\tSending delay in milliseconds (default 1ms)\n");
		printf("\t-c <alg>\tSpecify encryption algorithm to use\n\n");
		printf("Supported Algorithms:\n\t");
		for(i = 1; i <= CAMELOID_ALGO_COUNT;i++)
		{
			if(i != CAMELOID_ALGO_COUNT)
			printf("%s, ",cameloid_algo_to_name(i));
			else
			printf("%s ",cameloid_algo_to_name(i));
		}
		
		printf("\n\n");
		return 0;
	}

	while((c = getopt(argc,argv,"u:p:d:c:r:w:hvsD:")) != -1)
	{
		switch(c)
		{

			case 'c':
				choice_algo = cameloid_name_to_algo(optarg);
				break;

			case 'r':
				sc_read_dev = optarg;
				break;

			case 'w':
				sc_write_dev = optarg;
				break;

			case 'u':
				username  = optarg;
				break;

			case 'p':
				target_port = atoi(optarg);
				break;

			case 'h':
				card_duplex = SOUND_CARD_HALF;
				break;

			case 'd':
				data_port  =  atoi(optarg);
				break;

			case 'v':
				verbose = 1;
				break;

			case 's':
				use_strong_rand = 1;
				break;

			case 'D':
				send_delay = atoi(optarg);
				if(send_delay < 0)
					send_delay = 1;				
			
				break;
		}

	}


	hp = argv[optind];
	if(!hp)
	{
		printf("Please specify a host to call\n\n");
		return 0;
	}

	if((target_port < 0 || target_port > 65535))
	{
		printf("Invalid target port specified\n\n");
		return 0;
	}

	if(!target_port)
	{
		target_port = CAMELOID_LISTEN_PORT;
	}

	if((data_port  < 0 || data_port  > 65535))
	{
		printf("Invalid data port specified\n\n");
		return 0;
	}
	
	if(!data_port)
	{
		data_port = CAMELOID_DATA_PORT;
	}

	if(!username)
	{
		printf("Please set your username with the '-u' option\n\n");
		return 0;
	}

	if(strlen(username) > 10)
	{
		printf("Username is too long\n\n");
		return 0;
	}

	if(!(target_host = gethostbyname(hp)))
	{
		cameloid_print_error("Invalid host name specified");
		return 0;
	}


	memcpy(&addr,target_host->h_addr,4);
	
	printf("Initializing\n");

	if(choice_algo<0)
	{
		printf("Invalid encryption algorithm specified.\n\n");
		return 0;
	}

	if(!choice_algo)
		choice_algo = cameloid_get_random_algo();

	setuid(0);
	cameloid_lock_all_memory();
	if((cameloid_check_sound_card_lock())<0)
	{
		printf("Soundcard is busy, cannot continue.\n");
		return 0;
	}
	else
	{
		cameloid_lock_sound_card();
		my_lock = 1;
		atexit(cameloid_unlock_sound_card);
	}

	/* Lets setup some signals */
	signal(SIGQUIT, kill_connection);
        signal(SIGINT,	kill_connection);
        signal(SIGTERM,	kill_connection);
        signal(SIGSEGV,	kill_connection);
        signal(SIGHUP,	kill_connection);
        signal(SIGBUS,	kill_connection);
        signal(SIGFPE,	kill_connection);
        signal(SIGPIPE,	kill_connection);
	signal(SIGTSTP, half_duplex_switch);
	signal(SIGCHLD,	SIG_IGN);

	if((cameloid_allocate_memory_struct(&memctx))<0)
		return -1;
	else
	{
		SK = memctx.SK;
		NK = memctx.NK;
	}

	if(!card_duplex)
	{
		if((card_duplex = cameloid_soundcard_mode(sc_read_dev))<0)
		{
			kill_connection(0);
			return 0;
		}
	}


	switch(card_duplex)
	{
		case SOUND_CARD_FULL:
			printf("Sound card type FULL duplex\n");
			break;

		case SOUND_CARD_HALF:

			printf("Sound card type HALF duplex\n");
			break;
	
	}


	printf("Awaiting Negotiation-Key (NK) to initiate session\n");
	if((cameloid_get_nk(stdin,&memctx))<0)
	{
		printf("Too many failed attempts at entering NK.\n\n");
		cameloid_unlock_sound_card();
		return 0;
	}

	printf("Opening random devices %s and %s\n",RANDOM_DEVICE,URANDOM_DEVICE);	
	randomfd 	= cameloid_open_random(); 
	urandomfd 	= cameloid_open_urandom();


	cameloid_open_conn(addr,username);	
	cameloid_unlock_sound_card();
	printf("\n");
	return 0;

}
