/*************************************************
 * CAMELOID Secure End-to-End Telephony System
 * Copyright (C) 2005  Dr. Tek                     
 ************************************************/

#include "cameloid.h"
#include "cameloid_memory.h"
#include "cameloid_voice.h"
#include "cameloid_protocol.h"
#include "cameloid_functions.h"



/* Globals */
int 			randomfd, urandomfd, server_socket,
			soundfd, serv_port;

cameloid_kc_t 		*keystruct = NULL;
cameloid_mstruct_t	 memctx;

u8 			*sc_read_dev  = CAMELOID_SOUND_DEVICE, 
			*sc_write_dev = CAMELOID_SOUND_DEVICE;

pthread_t 		pthread_st;


__inline__
void kill_destroy_buffers(void)
{

	cameloid_destroy_memory(SK,CAMELOID_MAX_SK);
        cameloid_destroy_memory(NK,CAMELOID_MAX_NK);	

	if(keystruct)
                cameloid_destroy_memory((u8 *) keystruct, sizeof(cameloid_kc_t));

}

__inline__
void kill_destroy_sockets(void)
{
	if(soundfd > 0)
	{
		close(soundfd);
		soundfd = -1;
	}

	if(active_socket)
	{
		close(active_socket);
		active_socket = 0;
	}


}

__inline__
void kill_unlock_sc(void)
{
	if(my_lock)
	{
		cameloid_unlock_sound_card();
		my_lock = 0;
	}
}

__inline__
void kill_connection(int sig)
{
	struct termios termset;

	kill_destroy_buffers();
	kill_unlock_sc();

	/* This is done to ensure that the terminal will
           have the echo flags restored in case of a pre-mature
           cancellation */ 

         tcgetattr(fileno(stdin),&termset);
         termset.c_lflag |= (ECHO | ECHOE | ECHOK | ECHONL);
         tcsetattr(fileno(stdin),TCSANOW,&termset);

	
	if(sig != SIGSEGV)
	{
		if(active_socket || sig == SIGPIPE)
		{
			if(active_socket) 
			{
				printf("\nConnection Terminated\n"); fflush(stdout);
				kill_destroy_sockets();
				active_socket = 0;
			}
			return;
		}
	}
        else
	{
		printf("Segmentation Fault Captured\n"); fflush(stdout);
	}

	kill_destroy_sockets();


	if(!active_socket && (sig == SIGINT || sig == SIGTERM || sig == SIGHUP || sig == SIGSEGV))
	{
		cameloid_destroy_memory_struct(&memctx);
		printf("Shutting down server\n\n");
		close(server_socket);
		close(randomfd);
		close(urandomfd);
		exit(0);
	}
}


__inline__
void close_active_socket(int sig)
{
	close(active_socket);

}

__inline__
void cameloid_wait(int s, int data_socket, int out_socket)
{
	int 		 	 sklen = sizeof(struct sockaddr_in), 
				 use_algo, pt, size, rsp, r_skt = 0, nduplex, err_code;
	u8 			 *buf, *nonce, *digest;
	u16 			 sender_data_port;
	struct sockaddr_in 	 sk;
	cameloid_auth_t 	*auth_h;
	cameloid_init_pkt_t 	*init_h;
	cameloid_cr_pkt_t 	*cr_h, snd_cr_h;
	cameloid_challenge_t 	 chal;
	cameloid_sk_xchg_t	 sk_h;
	cameloid_kc_t 		 ctx;
	fd_snd_t 		 fdsndst;
	static struct timeval	 tv;


	tv.tv_sec 		= CAMELOID_NET_TIMEOUT;
	tv.tv_usec		= 0;


	buf	= (u8 *) cameloid_alloc(RD_BUFFER_SIZE + 1);
	if(!buf)
	{
		return;
	}

	nonce	= (u8 *) cameloid_alloc(CAMELOID_NONCE_SIZE + 1);
	if(!nonce)
	{
		return;
	}

	digest	= (u8 *) cameloid_alloc(16); 
	if(!digest)
	{
		return;
	}

	keystruct = &ctx;
	while(1)
	{

		if(r_skt)
		{
			close(r_skt);
		}

		sender_data_port = use_algo = size = pt = rsp = r_skt = nduplex = err_code = my_lock = active_socket = 0;

		clear_reset_screen();
		cameloid_destroy_memory(digest,16);
		cameloid_destroy_memory(nonce,CAMELOID_NONCE_SIZE + 1);
		cameloid_destroy_memory(NK,CAMELOID_MAX_NK);
		cameloid_destroy_memory(SK,CAMELOID_MAX_SK);
		cameloid_destroy_memory((u8 *)&sk,sizeof(struct sockaddr_in));
		cameloid_destroy_memory(buf,RD_BUFFER_SIZE);
		cameloid_destroy_memory((u8 *) keystruct, sizeof(cameloid_kc_t));

		printf("Awaiting secure voice communications\n");
	
		if((listen(s,1))<0)
		{
			continue;
		}
	
		if((r_skt = accept(s,(struct sockaddr *) &sk,&sklen))<0)
		{
			continue;
		}

		setsockopt(r_skt,SOL_SOCKET,SO_RCVTIMEO,&tv,sizeof(struct timeval));
		if((cameloid_check_sound_card_lock()) < 0)
		{
			close(r_skt);
			continue;
		}

		if((size = cameloid_net_read(r_skt,buf,PKT_IN_SIZE)) != PKT_IN_SIZE)
		{
			close(r_skt);
			continue;
		}
					
		auth_h = (cameloid_auth_t *) buf;
		pt   = auth_h->pkt_type;
	
		if(pt != CAMELOID_PKT_INIT)
		{
			close(r_skt);
			continue;
		}

		if(strncmp(auth_h->proto_id,CAMELOID_PROTO_IDENT,4))
		{
			close(r_skt);
			continue;
		}	
		else
		{   	
			nduplex = auth_h->duplex;
			if(nduplex != card_duplex) 
			{
				nduplex = SOUND_CARD_HALF;
			}
			
			init_h   = (cameloid_init_pkt_t *) (buf);
			use_algo = init_h->alg;

 			if(!cameloid_algo_to_name(use_algo))
			{
				close(r_skt);
				continue;
			}

			cameloid_play_ring(sc_write_dev);
			
			/* Check to see if username is obfuscated by looking at the initial nonce */
			/* older clients have the nonce set to 32 bytes of 0x30 (0) chars */
  
			if(memcmp(init_h->ath.nonce,"00000000000000000000000000000000",32) != 0)
			{
				/* De-obfuscate the username if it was obfuscated */ 
				cameloid_add_gf256(init_h->username,init_h->ath.nonce,CAMELOID_MAX_USERNAME);
			}

			printf("User %s from %s wants to establish a secure connection\nThe algorithm chosen was: %s\n",init_h->username,inet_ntoa(sk.sin_addr),cameloid_algo_to_name(use_algo));	
			printf("Do you want to accept the call [Y]es [N]o ? "); fflush(stdout);
			
			do_system_beep();
					
			while(rsp != 'Y' && rsp != 'N')
			{
				rsp = toupper(getchar());
			}
									
			if(rsp == 'N')
			{
				printf("Hanging up on caller %s\n",init_h->username);
				close(r_skt);
				continue;
			}

			if(rsp == 'Y')
			{
				active_socket 	 = r_skt;
				use_algo 	 = init_h->alg;
				sender_data_port = init_h->data_port;
			
				getchar();
	                        if((cameloid_get_nk(stdin,&memctx))<0)
				{
					printf("Too many failed NK attempts.\n");
					fflush(stdout);
					sleep(1);
					kill_connection(0);
					continue;
				}
								
        	                cameloid_key_setup(use_algo,NK,strlen(NK),&ctx);
				cameloid_destroy_memory(NK,strlen(NK));
				printf("Initiating new secure session with %s\n",inet_ntoa(sk.sin_addr)); 
				printf("Generating random challenge\n");

				/* Generate our random challenge */			
				cameloid_get_p_random_bytes(randomfd,chal.chal,32);
				cameloid_get_p_random_bytes(urandomfd,nonce,CAMELOID_NONCE_SIZE);
			        cameloid_generate_hash(chal.chal,32,chal.hash);
   
        			cameloid_crypt_data(CAMELOID_MODE_ENCRYPT,
						    chal.chal,
						    32,
						    nonce,&ctx);	
				
				printf("Sending random challenge and awaiting response\n"); 
				cameloid_build_cr_packet(&snd_cr_h,chal.chal,nonce,CAMELOID_PKT_CHALLENGE);
				snd_cr_h.ath.duplex = nduplex;

				cameloid_hmac_md5(snd_cr_h.data,32,nonce,CAMELOID_NONCE_SIZE,
						  ctx.keydatahash,CAMELOID_MAC_KEYSIZE,snd_cr_h.ath.hmac);

				if((send(r_skt,&snd_cr_h,PKT_CR_SIZE,0))<0)
				{
					kill_connection(0);
					continue;
				}

				if((cameloid_net_read(r_skt,buf,PKT_CR_SIZE)) != PKT_CR_SIZE)
				{
					kill_connection(0);
					continue;
				}
				
				cr_h = (cameloid_cr_pkt_t *) buf;				
				auth_h = &cr_h->ath;
				
				if(auth_h->pkt_type != CAMELOID_PKT_RESPONSE)
				{
					cameloid_print_error("Unexpected packet received.");
					kill_connection(0);
					continue;
				}
				else
				{
					cameloid_hmac_md5(cr_h->data,
							  32,
							  auth_h->nonce,
							  CAMELOID_NONCE_SIZE,
							  ctx.keydatahash,
							  CAMELOID_MAC_KEYSIZE,
							  digest);
			
					if(memcmp(digest,auth_h->hmac,16))
					{
						 cameloid_print_error("*Possible* injection attack, Invalid MAC computed for C/R packet.");
						 kill_connection(0);
						 continue;
					} 

					cameloid_crypt_data(CAMELOID_MODE_DECRYPT,
							    cr_h->data,
							    32,
							    auth_h->nonce,&ctx);
				
					if(memcmp(cr_h->data,chal.hash,32))
					{
						printf("Invalid response to challenge\n");
						kill_connection(0);
						continue;
					}
					else
					{
						printf("Response to challenge is valid\n");
						printf("Generating %3d-bit random SK\n",CAMELOID_MAX_SK_SEG*8);
						cameloid_get_r_random_bytes(randomfd,SK,CAMELOID_MAX_SK_SEG);

						printf("Generated SK segment hash -> %32s\n",cameloid_hash_and_show_md5(SK,CAMELOID_MAX_SK_SEG));
						cameloid_get_p_random_bytes(urandomfd,nonce,CAMELOID_NONCE_SIZE);

						memcpy(sk_h.sk_segment,SK,CAMELOID_MAX_SK_SEG);
						cameloid_crypt_data(CAMELOID_MODE_ENCRYPT,
								    sk_h.sk_segment,
								    CAMELOID_MAX_SK_SEG,
								    nonce,&ctx);
					
						cameloid_build_auth_packet(&sk_h.ath,nonce,CAMELOID_PKT_SK);																
						
						cameloid_hmac_md5(sk_h.sk_segment,
								  32, 
								  nonce, 
								  CAMELOID_NONCE_SIZE,
								  ctx.keydatahash,
                                                  		  CAMELOID_MAC_KEYSIZE,
								  sk_h.ath.hmac);

						if((send(r_skt,&sk_h,PKT_SK_SIZE,0)) != PKT_SK_SIZE)
						{
							kill_connection(0);
							continue;
						}
					
						printf("Awaiting SK segment from %s\n",inet_ntoa(sk.sin_addr));
					
						if((cameloid_net_read(r_skt,(u8 *) &sk_h,PKT_SK_SIZE)) != PKT_SK_SIZE)
						{
							kill_connection(0);
							continue;
						}
						
						if(sk_h.ath.pkt_type == CAMELOID_PKT_SK)
						{
							 cameloid_hmac_md5(sk_h.sk_segment,
									   32,
									   sk_h.ath.nonce,
									   CAMELOID_NONCE_SIZE,
                                          				   ctx.keydatahash,
									   CAMELOID_MAC_KEYSIZE,
									   digest);
                        
                              		  		if(memcmp(sk_h.ath.hmac,digest,16))
                                			{
                                	       	 		printf("Possible injection attack, Invalid MAC computed for C/R packet.\n"); fflush(stdout);
                        	                		kill_connection(0);
                	                        		continue;
		                                	}

							printf("Session key segment received\n");
							cameloid_crypt_data(CAMELOID_MODE_DECRYPT,
									    sk_h.sk_segment,
									    CAMELOID_MAX_SK_SEG,
									    sk_h.ath.nonce,&ctx);

							/* Destroy old keychain */
							cameloid_destroy_memory((u8 *) &ctx,sizeof(cameloid_kc_t));
						
							printf("Received SK segment hash -> %32s\n",cameloid_hash_and_show_md5(sk_h.sk_segment,CAMELOID_MAX_SK_SEG));
							memcpy(SK+CAMELOID_MAX_SK_SEG,sk_h.sk_segment,CAMELOID_MAX_SK);
							printf("Session key hash -> %32s\n",cameloid_hash_and_show_md5(SK,CAMELOID_MAX_SK));


							/* Run newly created SK through the selected KSA */
							cameloid_key_setup(use_algo,SK,CAMELOID_MAX_SK,&ctx);

							/* Destroy SK buffer */
							cameloid_destroy_memory(SK,CAMELOID_MAX_SK);

							cameloid_lock_sound_card();
							my_lock = 1;

							close(r_skt);
							active_socket = 255;

							if(nduplex == SOUND_CARD_HALF)
							{
								printf("Communicating securely in HALF duplex mode\n"); fflush(stdout);

								while(active_socket)
								{

									printf("Secure Listening Mode          \r"); fflush(stdout);
									soundfd = cameloid_open_sound(sc_write_dev,O_WRONLY);
		                                                                
                                                        		if(soundfd<0)
                                                        		{
                                                               		 	cameloid_destroy_memory((u8 *) &ctx,sizeof(cameloid_kc_t));
                                                               	 		cameloid_destroy_memory((u8 *) SK, CAMELOID_MAX_SK);
                                                                  		close(r_skt);
                                                                  		continue;
                                                        		}
	
									if((cameloid_write_half_duplex(data_socket,soundfd,urandomfd,&ctx))<0)
									{
										kill_connection(0);	
										break;
									}
	
									cameloid_close_sound(soundfd);
									soundfd = cameloid_open_sound(sc_read_dev,O_RDONLY);
								
                                                 	               if(soundfd<0)
                                                        	        {
                                                               		        cameloid_destroy_memory((u8 *) &ctx,sizeof(cameloid_kc_t));
                                                                	        cameloid_destroy_memory((u8 *) SK, CAMELOID_MAX_SK);
                                                                        	continue;
                                                                	}

									printf("Secure Talking Mode      \r"); fflush(stdout);
									if((cameloid_read_half_duplex(out_socket,soundfd,urandomfd,sender_data_port,sk.sin_addr.s_addr,&ctx))<0)
									{
										kill_connection(0);
										break;
									}
									cameloid_close_sound(soundfd);
                                       				
							}
						}
						else
						{
		
							fdsndst.readsocket  	= data_socket;
							fdsndst.writesocket	= out_socket;
							fdsndst.urandomfd 	= urandomfd;
							fdsndst.ctx 	  	= &ctx;
							fdsndst.sender_port	= sender_data_port;
							fdsndst.sender_addr	= sk.sin_addr.s_addr;
		
							printf("Communicating securely in FULL duplex mode\n"); fflush(stdout);	
							if((soundfd = cameloid_open_sound(sc_read_dev,O_RDWR))<0)
							{
								kill_connection(0);
								continue;
							}
		
							fdsndst.soundfd  = soundfd;
							pthread_create(&pthread_st,NULL,(void *) cameloid_read_full_duplex,(void *) &fdsndst);
							
							cameloid_write_full_duplex((void *)&fdsndst);
							cameloid_unlock_sound_card();
							my_lock = 0;
							if(soundfd) 
							{
								close(soundfd);
								soundfd = 0;
							}	
							continue;

						}
				
						}		
						
							if(soundfd) 
							{
								close(soundfd);
								soundfd = 0;
							}
	
							my_lock = 0;
							cameloid_unlock_sound_card();
							continue;			
						}
												
				}
					
				}
																															
						
			}
						
		  }										
				  				
	
}

__inline__
void cameloid_bind_port(void)
{
	int s, data_socket, out_socket, v = 1, bufsize = 40000;
	struct sockaddr_in sk;
	struct timeval tv;

	if((s 			= socket(AF_INET,SOCK_STREAM,0))<0)
	{
		cameloid_print_error("Cannot open TCP network socket.");
		return;
	}

	if((data_socket 	= socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP))<0)
	{
		cameloid_print_error("Cannot open UDP network socket.");
		return;
	}

	if((out_socket		= dup(data_socket))<0)
	{
		cameloid_print_error("Cannot duplicate UDP network socket.");
                return;
	}

	sk.sin_port 		= HTONS(serv_port);
	sk.sin_addr.s_addr 	= 0;
	sk.sin_family 		= AF_INET;
	
	setsockopt(s, SOL_SOCKET,SO_REUSEADDR, (char*)&v, sizeof(v));
	if((bind(s,(struct sockaddr *) &sk, sizeof(struct sockaddr_in)))<0)
	{
		printf("Cannot bind to TCP port %5d\n",serv_port);
		return;
	}

	sk.sin_port		= HTONS(data_port);
	setsockopt(data_socket, SOL_SOCKET,SO_REUSEADDR, (char*)&v, sizeof(v));
	setsockopt(out_socket,  SOL_SOCKET,SO_SNDBUF, &bufsize, sizeof(int));
	setsockopt(data_socket, SOL_SOCKET,SO_RCVBUF, &bufsize, sizeof(int));
	
	if((bind(data_socket,(struct sockaddr *) &sk, sizeof(struct sockaddr_in)))<0)
	{
		printf("Cannot bind to UDP port %5d\n",data_port);
		return;
	}

	tv.tv_sec 	= CAMELOID_NET_UDP_TIMEOUT;
	tv.tv_usec	= 0;
	setsockopt(data_socket,SOL_SOCKET,SO_RCVTIMEO,&tv,sizeof(struct timeval));
	
	server_socket = 0;
	cameloid_wait(s,data_socket,out_socket);	
	
}

int main(int argc, char **argv)
{
	int j = 0;

	close(2);       /* Close stderr because Speex is noisy */
	use_strong_rand = randomfd = urandomfd = card_duplex = server_socket = soundfd = 0;
	data_port = serv_port = my_lock = sound_switch = 0;

	printf("\n\tCAMELOID v%s\n\tSecure Network Telephony Server\n\n",VERSION);

	send_delay = 1;

	if(argc > 1)
	{
		while((j = getopt(argc,argv,"p:d:D:sh")) != -1)
		{


			switch(j)
			{
				case 'h':

					printf("usage: %s [options]\n\nOptions:\n",argv[0]);
					printf("\t-p[port]	Bind to an alternate server port (default: %5d)\n",CAMELOID_LISTEN_PORT); 
					printf("\t-d[port]	Bind to an alternate data port   (default: %5d)\n",CAMELOID_DATA_PORT);
					printf("\t-D <ms>	Sending delay in milliseconds (default 1ms)\n");
					printf("\t-s		Use strong random numbers (rijndael encrypted)\n");
					printf("\t-h		Show this help menu\n\n");
					return 0;
					break;

					
				case 's':
				
					use_strong_rand = 1;
					break;


				case 'D':

					send_delay = atoi(optarg);
					if(send_delay<0)
					{
						send_delay = 1;
					}
					break;

		
				case 'p':

					serv_port = atoi(optarg);
					if(serv_port<0 || serv_port > 65535)
					{
						printf("Invalid port specified\n");
						return 0;
					}
					break;	
			
				case 'd':
	
					serv_port = atoi(optarg);
                	                if(serv_port<0 || serv_port > 65535)
                        	        {
                               		        printf("Invalid port specified\n");
                                        	return 0;
                                	}
					break;

			}

		}

	}
	

	
	printf("Initializing\n");
	setuid(0);
	cameloid_lock_all_memory();	

	if((card_duplex = cameloid_soundcard_mode(SOUND_CARD_DEVICE))<0)
	{
		exit(0);
	}		
        
        switch(card_duplex)
        {
                case SOUND_CARD_FULL:
                        printf("Sound card type FULL Duplex\n");
                        break;
        
                case SOUND_CARD_HALF:
                        printf("Sound card type HALF Duplex\n");
                        break;

        }


	if((cameloid_allocate_memory_struct(&memctx))<0)
		return -1;
	else
	{
		NK 	= memctx.NK;
		SK 	= memctx.SK;
	}
	printf("Opening random device %s\n",RANDOM_DEVICE);
	
	randomfd  	= cameloid_open_random(); 
	urandomfd	= cameloid_open_urandom();


	if(!data_port)
	{
		data_port = CAMELOID_DATA_PORT;
	}

	if(!serv_port)
	{
		serv_port = CAMELOID_LISTEN_PORT;
	}

	signal(SIGINT,	kill_connection);
	signal(SIGSEGV,	kill_connection);  
	signal(SIGTERM,	kill_connection);
	signal(SIGHUP,	kill_connection);
	signal(SIGFPE,	kill_connection);
	signal(SIGBUS,	kill_connection);
	signal(SIGPIPE,	kill_connection);
	signal(SIGQUIT,	kill_connection);
	signal(SIGTSTP, half_duplex_switch);
	signal(SIGCHLD,	SIG_IGN);
	
	cameloid_bind_port();
	return 0;

}
