/*************************************************
 * CAMELOID Secure End-to-End Telephony System
 * Copyright (C) 2005  Dr. Tek                     
 ************************************************/


#include <netdb.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>


/*  Key Sizes */

#define		CAMELOID_MIN_NK				16
#define		CAMELOID_MAX_NK				256
#define		CAMELOID_MAX_SK				64
#define		CAMELOID_MAX_SK_SEG			32
#define		CAMELOID_NONCE_SIZE			32
#define		CAMELOID_MAC_KEYSIZE			10
#define		CAMELOID_PROTO_IDENT			"CVEP\0"

/* Packet Types */

#ifndef		VIDEO_CAP
#define		CAMELOID_PKT_CHALLENGE			0xEF
#define		CAMELOID_PKT_RESPONSE			0xF3
#define		CAMELOID_PKT_INIT			0xF9
#define		CAMELOID_PKT_DATA			0xF8
#define		CAMELOID_PKT_SWITCH			0xFF
#define		CAMELOID_PKT_SK				0xD0
#endif

#define		CAMELOID_PKT_PING			0x41
#define		CAMELOID_PKT_PONG			0x42

/* Video packet types */

#define		CAMELOID_PKT_V_INIT			0x01
#define		CAMELOID_PKT_V_CHALLENGE		0x02
#define		CAMELOID_PKT_V_RESPONSE			0x03
#define		CAMELOID_PKT_V_SK			0x04
#define		CAMELOID_PKT_V_DATA			0x05


/* Instant Messenger Packet Types */
#define		CAMELOID_PKT_IM_INIT			0x06
#define         CAMELOID_PKT_IM_CHALLENGE               0x07
#define         CAMELOID_PKT_IM_RESPONSE                0x08
#define         CAMELOID_PKT_IM_SK                      0x09
#define         CAMELOID_PKT_IM_DATA                    0x0A


#define		CAMELOID_IM_LISTEN_PORT			7405
#define		CAMELOID_IM_SERVER_PORT			7406

#define		CAMELOID_VIDEO_LISTEN_PORT		7403
#define		CAMELOID_VIDEO_DATA_PORT		7404

#define		CAMELOID_LISTEN_PORT			7401
#define		CAMELOID_DATA_PORT			7402

#define		CAMELOID_NET_TIMEOUT			40
#define		CAMELOID_NET_UDP_TIMEOUT		7

/* Error Packets */
#define		CAMELOID_ERROR_INVALID_CR		1


/* Packet Structures */

#ifndef	VIDEO_CAP
#ifndef	IM_CAP
typedef struct
{
	u8 proto_id[4], hmac[16], nonce[CAMELOID_NONCE_SIZE], 
	   duplex, pkt_type;
} cameloid_auth_t;

#endif
#endif


# ifdef VIDEO_CAP 
typedef struct
{
	u8 proto_id[4], 
	   hmac[16], 
	   nonce[CAMELOID_NONCE_SIZE], 
	   pkt_type;
} cameloid_auth_t;

#endif

#ifdef	IM_CAP
typedef	struct
{
	u8 proto_id[4],
	hmac[16],
	nonce[CAMELOID_NONCE_SIZE],
	pkt_type;
} cameloid_auth_t;

#endif

#ifdef	VIDEO_CAP
typedef	struct
{
	cameloid_auth_t ath; 
	u16 	vdlen;
} cameloid_vd_auth_t;


#define		PKT_VDA_SIZE	sizeof(cameloid_vd_auth_t)
#endif 


#ifdef	 IM_CAP
typedef struct
{
        cameloid_auth_t ath;
        u8 alg, username[10];
} cameloid_init_pkt_t;
#endif

#ifndef IM_CAP
typedef struct 
{
	cameloid_auth_t ath;
	u8 alg, username[10];
	u16 data_port;
} cameloid_init_pkt_t;
#endif

typedef struct
{
	cameloid_auth_t ath;
	u8 data[32];

} cameloid_cr_pkt_t;

typedef struct
{
	cameloid_auth_t ath;
	u8 sk_segment[32];

} cameloid_sk_xchg_t;

typedef struct 
{
	u8 chal[32], hash[32];
} cameloid_challenge_t; 

	
/* Packet Header Sizes */

#define		PKT_CR_SIZE		sizeof(cameloid_cr_pkt_t)
#define		PKT_IN_SIZE		sizeof(cameloid_init_pkt_t)
#define		PKT_AT_SIZE		sizeof(cameloid_auth_t)
#define		PKT_SK_SIZE		sizeof(cameloid_sk_xchg_t)
#define		PKT_DT_SIZE		200
#define		PKT_UC_SIZE		160


/* Endian-specific Byte-Ordering Routines */

#ifdef		LITTLE_ENDIAN_ARCH

#ifndef		HTONS
#define		HTONS(x)		htons(x)
#endif

#ifndef		HTONL
#define		HTONL(x)		htonl(x)
#endif

#ifndef         NTOHS
#define         NTOHS(x)                ntohs(x)
#endif

#ifndef         NTOHL
#define         NTOHL(x)                ntohl(x)
#endif

#else 

#ifdef		BIG_ENDIAN_ARCH	

#ifndef		HTONS
#define		HTONS(x)		(x)
#endif

#ifndef		NTOHS
#define		NTOHS(x)		(x)
#endif

#ifndef		NTOHL		
#define		NTOHL(x)		(x)
#endif


#ifndef		HTONL
#define		HTONL(x)		(x)
#endif

#endif
#endif

