/*************************************************
 * CAMELOID Secure End-to-End Telephony System
 * Copyright (C) 2005  Dr. Tek                     
 ************************************************/

#include "cameloid.h"
#include "cameloid_protocol.h"
#include "cameloid_functions.h"
#include "cameloid_voice.h"
#include <sys/stat.h>
#include <errno.h>
#include <speex/speex.h>

void 		*dec_state, *enc_state;
SpeexBits 	 ebits, dbits;

__inline__
void half_duplex_switch(int sig)
{
        sound_switch = 1;
}                       

__inline__
u32 cameloid_get_output_space(int fd)
{
	u32 		outsize, fsize;
	audio_buf_info  info;

	outsize = PKT_UC_SIZE * 2;
	cameloid_destroy_memory((u8 *)&info,sizeof(audio_buf_info));

	if((ioctl(fd,SNDCTL_DSP_GETOSPACE,&info))<0)
	{
		return outsize;
	}
	
	fsize = info.fragments * info.fragsize;
	if(fsize && (fsize < outsize))
	{
		outsize = fsize;
	}

	return outsize;
		
}


__inline__
u32 cameloid_get_input_space(int fd)
{
        u32 		outsize, fsize;
        audio_buf_info  info;

        outsize = PKT_UC_SIZE * 2;
        cameloid_destroy_memory((u8 *)&info,sizeof(audio_buf_info));
        if((ioctl(fd,SNDCTL_DSP_GETISPACE,&info))<0)
        {
	       return outsize;
	}
	
	fsize = info.fragments * info.fragsize;

        if(fsize && (fsize < outsize))
	{	
               outsize = fsize;
	}

        return outsize;

}


__inline__
void cameloid_close_sound(int fd)
{
	ioctl(fd,SNDCTL_DSP_SYNC,0);
	ioctl(fd,SNDCTL_DSP_RESET,0); 
	close(fd);
}

__inline__
int cameloid_configure_audio(int fd, int mode)
{
        u32 v = 0;

        v = (4 << 16) | 8;
        if(ioctl(fd,SNDCTL_DSP_SETFRAGMENT,&v))
        {
                cameloid_print_error("Error configuring sound");
                return -1;
        }
	
        v = AFMT_S16_LE;
        if(ioctl(fd,SNDCTL_DSP_SETFMT,&v))
        {
                cameloid_print_error("Error configuring sound");
                return -1;
        }
	 
         v = SOUND_MONO;
         if(ioctl(fd,SNDCTL_DSP_CHANNELS,&v))
         {
                cameloid_print_error("Error configuring sound");
                return -1;
         }

         v = SOUND_SPEED;
         if(ioctl(fd,SNDCTL_DSP_SPEED,&v))
         {
                cameloid_print_error("Error configuring sound");
                return -1;
         }

	 if(mode != O_WRONLY && mode != O_RDONLY)
	 {
       		 ioctl(fd,SNDCTL_DSP_SETDUPLEX,0);
	 }

	 return fd;
}

__inline__
int cameloid_soundcard_mode(u8 *device)
{
        register int fd = 0;
        u32 caps = 0;   

        if((fd = open(device,O_RDWR))<0)
        {
                cameloid_print_error("Cannot open sound device.\n");
                return -1;
        }

          
        if(ioctl(fd,SNDCTL_DSP_GETCAPS,&caps))
        {  
                cameloid_print_error("Cannot get sound capabilites.\n");
                return -1;
        }  

        close(fd);
 
        if(!(caps & DSP_CAP_DUPLEX))
                return SOUND_CARD_HALF;

        return SOUND_CARD_FULL;

}

__inline__
int cameloid_open_sound(u8 *device, int mode)
{
        register int fd = 0;

        if((fd = open(device,mode))<0)
        {
                cameloid_print_error("Cannot open sound device.\n");
                return -1;
        }

        return cameloid_configure_audio(fd,mode);
}


__inline__
void cameloid_play_ring(u8 *sound_dev)
{
       register int i, fd = 0, ringfd = 0;
       u8 *data =NULL;
       struct stat st;	
         
       if((ringfd = open(RING_SOUND_FILE,O_RDONLY))<0)
 	{
		return;
 	}

       if((fd = open(sound_dev,O_WRONLY))<0)
	{
		close(ringfd);
		return;
	}
       
       	if((fstat(ringfd,&st))<0)
	{
		return;
	}

        if((data = (u8 *) calloc(1,st.st_size))<0)
	{
		return;
	}

	read(ringfd,data,st.st_size);
	close(ringfd);

	for(i = 2; i > 0; --i)
	{
		write(fd,data,st.st_size);
		sleep(1);
	}

	ioctl(fd,SNDCTL_DSP_RESET,0);
	close(fd);

	free(data); 
  	return;
}

__inline__
int cameloid_check_sound_card_lock(void)
{
	struct stat st;

	if((lstat(LOCK_FILE,&st)) == 0)
		return -1;

	return 0;
}
	


__inline__
int cameloid_lock_sound_card(void)
{
	register int lockfd = 0;

	if((lockfd = open(LOCK_FILE,O_WRONLY|O_CREAT|O_TRUNC,0666))<0)
	{
		cameloid_print_error("Cannot lock sound card");
		return -1;

	}	
	close(lockfd);
	return 0;
}

__inline__
void cameloid_unlock_sound_card(void)
{
	(void) unlink(LOCK_FILE);
}

__inline__
void cameloid_speex_encoder_init(void)
{
        int tmpv = 0;
        static int speex_init_done = 0;
        
	if(!speex_init_done)
        {
		enc_state = speex_encoder_init(&speex_nb_mode);
        	speex_bits_init(&ebits);
                                    
       	 	tmpv = 1;
        	speex_encoder_ctl(enc_state,SPEEX_SET_VAD,&tmpv);
        	speex_encoder_ctl(enc_state,SPEEX_SET_DTX,&tmpv);
        
        	tmpv = 8;
        	speex_encoder_ctl(enc_state,SPEEX_SET_QUALITY,&tmpv);

        	tmpv = SOUND_SPEED;
        	speex_encoder_ctl(enc_state,SPEEX_SET_SAMPLING_RATE, &tmpv);

        	tmpv = 2;
        	speex_encoder_ctl(enc_state,SPEEX_SET_COMPLEXITY,&tmpv);
		speex_init_done = 1;
	}
}


__inline__
void cameloid_speex_decoder_init(void)
{
        int tmpv = 0;
        static int speex_init_done = 0;
        
	if(!speex_init_done)
	{
       	 	dec_state = speex_decoder_init(&speex_nb_mode);       
        	speex_bits_init(&dbits);

        	tmpv = SOUND_SPEED;
        	speex_decoder_ctl(dec_state,SPEEX_SET_SAMPLING_RATE, &tmpv);

       		tmpv = 1;
        	speex_decoder_ctl(dec_state,SPEEX_SET_ENH,&tmpv);
 		
		speex_init_done = 1;
	}
}


__inline__
int cameloid_read_half_duplex(int skt, int sndfd, int urfd, u16 sender_port, u32 sender_addr, cameloid_kc_t *pctx)
{
	int err = 0;	
	u32 len, ocounter = 0, oseq_counter = 0;
	u8 nonce[CAMELOID_NONCE_SIZE+1];
	u8 data[PKT_DT_SIZE + PKT_AT_SIZE + 8 + 1];
	cameloid_auth_t *auth_h;
	short snd_data[PKT_UC_SIZE];
	float snd_databuf[PKT_UC_SIZE];
	cameloid_kc_t ctx;
	struct sockaddr_in sk;
	register u32 counter = 0, seq_counter = 0, i;
	struct timespec tv;

	cameloid_speex_encoder_init();


	tv.tv_sec       = 0;
        tv.tv_nsec      = (send_delay * 1000000);

	memcpy(&ctx,pctx,sizeof(cameloid_kc_t));
	sound_switch = 0;

	auth_h = (cameloid_auth_t *) data;
	
	cameloid_destroy_memory((u8 *)&sk,sizeof(struct sockaddr_in));	
	sk.sin_addr.s_addr = sender_addr;
	sk.sin_port	   = sender_port;
	sk.sin_family	   = AF_INET;

	/* Put UDP socket into connected state so we can receive ICMP errors */
	connect(skt,&sk,sizeof(sk));

	while(!sound_switch && active_socket)
	{
		cameloid_get_p_random_bytes(urfd,nonce,CAMELOID_NONCE_SIZE);
		cameloid_build_auth_packet(auth_h,nonce,CAMELOID_PKT_DATA);

		len = cameloid_get_input_space(sndfd);		
		if((err = read(sndfd,snd_data,len)) < len)
		{	
			if(err < 0)
			{
				cameloid_destroy_memory((u8 *) &ctx,sizeof(cameloid_kc_t));
				return -1;
			}
			break;
		}
			

		for( i = PKT_UC_SIZE; i > 0; --i)
		{
                        snd_databuf[i-1] = (float) snd_data[i-1];
		}

		speex_bits_reset(&ebits);
		speex_encode(enc_state,snd_databuf,&ebits);
		
		speex_bits_insert_terminator(&ebits);
		speex_bits_write(&ebits,data + PKT_AT_SIZE,PKT_DT_SIZE);
		
		cameloid_crypt_data(CAMELOID_MODE_ENCRYPT,
				   data + PKT_AT_SIZE,
				   PKT_DT_SIZE,
				   nonce,&ctx);

		counter++;
		if(!counter)
                {
                        seq_counter++;
                        oseq_counter = HTONL(seq_counter);
                        
                }

                ocounter 	= HTONL(counter);
	
                memcpy(data + (PKT_AT_SIZE + PKT_DT_SIZE) + 0,&ocounter,	4);
		memcpy(data + (PKT_AT_SIZE + PKT_DT_SIZE) + 4,&oseq_counter, 	4);

		cameloid_hmac_md5(data + PKT_AT_SIZE,
				  PKT_DT_SIZE + 8,
				  nonce,
				  CAMELOID_NONCE_SIZE,
                                  ctx.keydatahash,
				  CAMELOID_MAC_KEYSIZE,auth_h->hmac);

		if((send(skt,data,PKT_DT_SIZE + PKT_AT_SIZE + 8,0))<0)
		{
			cameloid_destroy_memory((u8 *) &ctx,sizeof(cameloid_kc_t));
			return -1;
		}
			
		nanosleep(&tv,NULL);
		
	}		

	counter++;
        if(!counter)
        {       
             	seq_counter++;
             	oseq_counter 	= HTONL(seq_counter);
 	}


	/* If we are here, send a switch packet */

        ocounter	= HTONL(counter);
	memcpy(data + (PKT_AT_SIZE + PKT_DT_SIZE) + 0,&ocounter,        4);
        memcpy(data + (PKT_AT_SIZE + PKT_DT_SIZE) + 4,&oseq_counter,    4);
    
	cameloid_get_p_random_bytes(urfd,nonce,32);
	cameloid_get_p_random_bytes(urfd,data + PKT_AT_SIZE,PKT_DT_SIZE);
	cameloid_build_auth_packet(auth_h,nonce,CAMELOID_PKT_SWITCH);

	cameloid_hmac_md5(data +  PKT_AT_SIZE,
                                  PKT_DT_SIZE + 8,   
                                  nonce,
                                  CAMELOID_NONCE_SIZE,
                                  ctx.keydatahash,
                                  CAMELOID_MAC_KEYSIZE,auth_h->hmac);

	cameloid_destroy_memory((u8 *) &ctx,sizeof(cameloid_kc_t));
	sendto(skt,data,PKT_DT_SIZE + PKT_AT_SIZE + 8,0,(struct sockaddr *) &sk,sizeof(struct sockaddr_in));	
	return 0;
	

}


__inline__
int cameloid_write_half_duplex(int skt, int sndfd, int urfd, cameloid_kc_t *pctx)
{
	u32 len, counter = 0, ocounter = 0, seq_counter = 0, oseq_counter = 0;
	int rsize;
        u8  data[PKT_DT_SIZE + PKT_AT_SIZE + 8 + 1];
        u8  digest[16];
	cameloid_auth_t *auth_h;
        short snd_data[PKT_UC_SIZE];
	float snd_databuf[PKT_UC_SIZE];
	cameloid_kc_t ctx; 
	register u32 i, last_counter = 0, last_seq_counter = 0;
	

	cameloid_speex_decoder_init();
        memcpy(&ctx,pctx,sizeof(cameloid_kc_t));
        auth_h = (cameloid_auth_t *) data;

        while(active_socket)
        {      
		
                if((rsize = cameloid_net_read_udp(skt,data,PKT_DT_SIZE+PKT_AT_SIZE + 8))<=0)
		{
			cameloid_destroy_memory((u8 *) &ctx,sizeof(cameloid_kc_t));
        		cameloid_destroy_memory(data,PKT_DT_SIZE + PKT_AT_SIZE + 1);
			return -1;
		}
			
		if(rsize < (PKT_DT_SIZE+PKT_AT_SIZE+8))
		{
			continue;
		}

		
		if(auth_h->pkt_type != CAMELOID_PKT_DATA && 
		   auth_h->pkt_type != CAMELOID_PKT_SWITCH)
		{
			continue;
		}

		cameloid_hmac_md5(data + PKT_AT_SIZE,
				  PKT_DT_SIZE + 8,
				  auth_h->nonce,
				  CAMELOID_NONCE_SIZE,
                                  ctx.keydatahash,
				  CAMELOID_MAC_KEYSIZE,digest);

		if((memcmp(digest,auth_h->hmac,16)))
		{
			continue;
		}
		
		memcpy(&ocounter,	data + (PKT_AT_SIZE + PKT_DT_SIZE) + 0, 4);
		memcpy(&oseq_counter,	data + (PKT_AT_SIZE + PKT_DT_SIZE) + 4, 4);

		
		counter 	= NTOHL(ocounter);
		seq_counter	= NTOHL(oseq_counter);

		if(counter <= last_counter || seq_counter < last_seq_counter )
		{
			continue;
		}

		if(auth_h->pkt_type == CAMELOID_PKT_SWITCH)
                {
                        break;
                }

		last_counter 		= counter;
		last_seq_counter	= seq_counter;

		cameloid_crypt_data(CAMELOID_MODE_DECRYPT,
				    data + PKT_AT_SIZE,
				    PKT_DT_SIZE,
				    auth_h->nonce,&ctx);
 
		cameloid_destroy_memory((u8 *)snd_data,PKT_UC_SIZE * 2);
		speex_bits_reset(&dbits);
		speex_bits_read_from(&dbits,(u8 *) data + PKT_AT_SIZE,PKT_DT_SIZE);
		
		if((speex_decode(dec_state,&dbits,snd_databuf))<0)
			continue;
               
		for(i = PKT_UC_SIZE; i > 0; --i)
                {
                        if(snd_databuf[i-1] >  32767.f) snd_databuf[i-1] =  32767.f;
                        if(snd_databuf[i-1] < -32768.f) snd_databuf[i-1] = -32768.f;
                        snd_data[i-1] = (short) snd_databuf[i-1];
                }

		len = cameloid_get_output_space(sndfd);
		write(sndfd,snd_data,len);
	
        }
 
	cameloid_destroy_memory((u8 *) &ctx,sizeof(cameloid_kc_t));
        cameloid_destroy_memory(data,PKT_DT_SIZE + PKT_AT_SIZE + 1);
        return 0;

}




__inline__
void cameloid_read_full_duplex(void *struct_ptr)
{
	u32 len, ocounter = 0, oseq_counter = 0;
	fd_snd_t *fdstruct;
	u8 nonce[CAMELOID_NONCE_SIZE+1], data[PKT_DT_SIZE + PKT_AT_SIZE + 8 + 1];
        cameloid_auth_t *auth_h;
        cameloid_kc_t ctx;
	s16 snd_data[PKT_UC_SIZE];
	float snd_databuf[PKT_UC_SIZE];
	struct sockaddr_in sk;
	int soundreadlen = 0;
	struct timespec tv;
	register u32 i, counter = 0, seq_counter = 0;

	cameloid_speex_encoder_init();


	fdstruct = (fd_snd_t *) struct_ptr;
        memcpy(&ctx,fdstruct->ctx,sizeof(cameloid_kc_t));

	auth_h = (cameloid_auth_t *) data;
	pthread_detach(pthread_self());


	cameloid_destroy_memory((u8 *)&sk,sizeof(struct sockaddr_in));
	sk.sin_port		= fdstruct->sender_port;
	sk.sin_addr.s_addr	= fdstruct->sender_addr;
	sk.sin_family		= AF_INET;

	tv.tv_sec 	= 0;
	tv.tv_nsec	= (send_delay * 1000000);

	while(active_socket)
	{
	
		len = cameloid_get_input_space(fdstruct->soundfd);

		if((soundreadlen = read(fdstruct->soundfd,snd_data,len))< len)
		 {
			if(soundreadlen < 0)
			{
				break;	
			}
			continue;
   		}
	
		for( i = PKT_UC_SIZE; i > 0; --i) 
		{
			snd_databuf[i-1] = (float) snd_data[i-1];
		}

		speex_bits_reset(&ebits);
		speex_encode(enc_state,snd_databuf,&ebits);

        	speex_bits_insert_terminator(&ebits);
		speex_bits_write(&ebits,data + PKT_AT_SIZE,PKT_DT_SIZE);    	
			
	
                cameloid_get_p_random_bytes(fdstruct->urandomfd,nonce,CAMELOID_NONCE_SIZE);
                cameloid_build_auth_packet(auth_h,nonce,CAMELOID_PKT_DATA);
		
	        cameloid_crypt_data(CAMELOID_MODE_ENCRYPT,
				    data + PKT_AT_SIZE,
				    PKT_DT_SIZE,
				    nonce,&ctx);

		counter++;
		if(!counter)
		{ 
			seq_counter++;
			oseq_counter = HTONL(seq_counter);
						
		}

		ocounter = HTONL(counter);

		memcpy(data + (PKT_AT_SIZE + PKT_DT_SIZE) + 0,&ocounter,4);
		memcpy(data + (PKT_AT_SIZE + PKT_DT_SIZE) + 4,&oseq_counter,4);

		cameloid_hmac_md5(data + PKT_AT_SIZE,
				  PKT_DT_SIZE + 8,
				  nonce,
				  CAMELOID_NONCE_SIZE,
                                  ctx.keydatahash,
				  CAMELOID_MAC_KEYSIZE,auth_h->hmac);

                sendto(fdstruct->writesocket,data,PKT_DT_SIZE + PKT_AT_SIZE + 8,MSG_DONTWAIT,
		       (struct sockaddr *) &sk,sizeof(struct sockaddr));
		
	
	
		nanosleep(&tv,NULL);
	}

	cameloid_destroy_memory((u8 *) &ctx,sizeof(cameloid_kc_t));
	return;
	
}


__inline__
void cameloid_write_full_duplex(void *struct_ptr)
{
	u32 len, counter = 0, ocounter = 0, oseq_counter = 0, seq_counter = 0;
	int rsize;
	u8  data[PKT_DT_SIZE + PKT_AT_SIZE + 8 + 1];
	u8  digest[16];
	cameloid_kc_t ctx;
	cameloid_auth_t *auth_h;
	fd_snd_t *fdstruct;
	short snd_data[PKT_UC_SIZE];
	float snd_databuf[PKT_UC_SIZE];
	register u32 i, last_counter = 0, last_seq_counter = 0;

	fdstruct 	= (fd_snd_t *) struct_ptr;
        auth_h		= (cameloid_auth_t *) data;

	memcpy(&ctx,fdstruct->ctx,sizeof(cameloid_kc_t));
	cameloid_speex_decoder_init();
	
        while(active_socket)
        {
	  
                if((rsize = cameloid_net_read_udp(fdstruct->readsocket,data,PKT_DT_SIZE+PKT_AT_SIZE + 8)) <= 0)
		{
			break;
		}

		if(rsize < (PKT_DT_SIZE+PKT_AT_SIZE+8))
		{
			continue;
		}

		if(auth_h->pkt_type != CAMELOID_PKT_DATA)
		{
			continue;
		}

		cameloid_hmac_md5(data + PKT_AT_SIZE,
				  PKT_DT_SIZE + 8,
				  auth_h->nonce,
				  CAMELOID_NONCE_SIZE,
                                  ctx.keydatahash,
				  CAMELOID_MAC_KEYSIZE,digest);

		if((memcmp(digest,auth_h->hmac,16))) 
		{
			continue;
		}

		memcpy(&ocounter,data+(PKT_AT_SIZE+PKT_DT_SIZE),4);
		memcpy(&oseq_counter, data + (PKT_AT_SIZE + PKT_DT_SIZE) + 4, 4);

                counter 		= NTOHL(ocounter);
		seq_counter		= NTOHL(oseq_counter);
		
		if(counter <= last_counter || seq_counter < last_seq_counter)
			continue; 
		
		last_counter 		= counter;	
		last_seq_counter	= seq_counter;

                cameloid_crypt_data(CAMELOID_MODE_DECRYPT,
				    data + PKT_AT_SIZE,
				    PKT_DT_SIZE,
				    auth_h->nonce,&ctx); 

		cameloid_destroy_memory((u8 *) snd_data,PKT_UC_SIZE * 2);
		speex_bits_read_from(&dbits,(u8 *) data + PKT_AT_SIZE,PKT_DT_SIZE);
	
		if((speex_decode(dec_state,&dbits,snd_databuf))<0)
		{
			continue;
		}
	
		for(i = PKT_UC_SIZE; i > 0; --i) 
		{
			if(snd_databuf[i-1] >  32767.f) snd_databuf[i-1] =  32767.f;
			if(snd_databuf[i-1] < -32768.f) snd_databuf[i-1] = -32768.f;
			snd_data[i-1] = (short) snd_databuf[i-1];		
		}
		
		len = cameloid_get_output_space(fdstruct->soundfd);
		write(fdstruct->soundfd,snd_data,len);

        }

	cameloid_destroy_memory((u8 *)&ctx,sizeof(cameloid_kc_t));
        cameloid_destroy_memory(data,PKT_DT_SIZE + PKT_AT_SIZE + 1);
        return;
}	
