/*************************************************
 * CAMELOID Secure End-to-End VideoPhone System                   
 * Copyright (C) 2005  Dr. Tek
 ************************************************/

#ifndef         VIDEO_CAP
#define         VIDEO_CAP
#endif

#ifndef		LINUX
#error		CAMELOID video currently works on linux only.. Sorry :(
#endif

#include "cameloid.h"
#include "cameloid_memory.h"
#include "cameloid_voice.h"
#include "cameloid_protocol.h"
#include "cameloid_functions.h"
#include <gtk/gtk.h>
#include "interface.h"
#include "support.h"

/* Globals */
int 			 choice_algo, randomfd, urandomfd,
    			 s, target_port = CAMELOID_VIDEO_LISTEN_PORT, 
			 data_port = CAMELOID_VIDEO_DATA_PORT;


cameloid_kc_t  		*keystruct;
cameloid_mstruct_t 	 memctx;

#define			 VIDBUFSIZE 		(40 * 1024)

__inline__
void daemonize()
{
	
	if(fork())
	{
		exit(0);
	}
}


__inline__
void kill_connection(int sig)
{
	struct termios termset;

	/* This is done to ensure that the terminal will 
	   have the echo flags restored in case of a pre-mature
	   cancellation */

	 tcgetattr(fileno(stdin),&termset);
	 termset.c_lflag |= (ECHO | ECHOE | ECHOK | ECHONL);
	 tcsetattr(fileno(stdin),TCSANOW,&termset);	
        

	printf("\nTerminating connection\n\n");
	cameloid_destroy_memory_struct(&memctx);
	
	if(keystruct)
	{
		cameloid_destroy_memory((u8 *) keystruct, sizeof(cameloid_kc_t));	
	}
	close(s);
	exit(0);
	
}

__inline__
int cameloid_open_conn(u32 addr, u8 *username)
{
	int 	 	 	 sklen = sizeof(struct sockaddr_in),  bufsize = VIDBUFSIZE,
			 	 s  = 0, data_socket = 0, cprecv = 0, rdlen = 0;

	struct sockaddr_in 	 sk, sk2, sk3;
	u8 			 *buf, *nonce,*digest, 
				 *ptr = NULL,  *hash, hmac[16], *databuf, *rgbptr;

	cameloid_init_pkt_t 	 init_h;
	cameloid_sk_xchg_t 	*sk_h, sk_w_h;
	cameloid_cr_pkt_t 	*cr, snd_cr_h;
	cameloid_auth_t 	*ath_w_p, *ath_r_p;
	cameloid_vd_auth_t	*vdptr, vdah;

	cameloid_kc_t 		 ctx;
	struct timeval	 	 tv;
	GtkWidget 		*mainwindow, *imageframe;
	u32			 oseq_counter, ocounter, counter,seq_counter, 
				 last_counter, last_seqcounter, macsize;

	GdkPixbuf		*pixbuf;


	buf	= (u8 *) cameloid_alloc(RD_BUFFER_SIZE + 1);
	if(!buf)
	{
		return 0;
	}

	hash 	= (u8 *) cameloid_alloc(32);
	if(!hash)
	{	
		return 0;
	}

	nonce	= (u8 *) cameloid_alloc(CAMELOID_NONCE_SIZE + 1);
	if(!nonce)
	{
		return 0;
	}

	digest	= (u8 *) cameloid_alloc(16); 
	if(!digest)
	{
		return 0;
	}
	
	cameloid_key_setup(choice_algo,NK,strlen(NK),&ctx);
	cameloid_destroy_memory(NK,strlen(NK));
	cameloid_destroy_memory((u8 *)&sk,  sklen);
	cameloid_destroy_memory((u8 *)&sk2, sklen);
	cameloid_destroy_memory((u8 *)&sk3, sklen);
	cameloid_destroy_memory((u8 *)&init_h, sizeof(init_h));

	keystruct = &ctx;

	last_counter = last_seqcounter = macsize = seq_counter = ocounter = oseq_counter = counter = 0;


	rgbptr = (u8 *) cameloid_alloc((320 * 240 * 3)+1);
	
	if(!rgbptr)
	{
		kill_connection(0);
	}

	
	if((data_socket = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP))<0)
        {                                                
                cameloid_print_error("Cannot open UDP network socket.");
                return -1;
        }

	tv.tv_sec        = CAMELOID_NET_UDP_TIMEOUT;
        tv.tv_usec       = 0; 


	setsockopt(data_socket, SOL_SOCKET,SO_RCVTIMEO,&tv,sizeof(struct timeval));                                                 
        setsockopt(data_socket, SOL_SOCKET,SO_RCVBUF,  &bufsize, sizeof(int));

        sk2.sin_port             = HTONS(data_port);
        sk2.sin_addr.s_addr      = addr;
        sk2.sin_family           = AF_INET;
	

	if((s = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP))<0)
	{
		cameloid_print_error("Error opening TCP socket.");
		return -1;

	}

	tv.tv_sec        = CAMELOID_NET_TIMEOUT;
        tv.tv_usec       = 0;

	setsockopt(s,SOL_SOCKET,SO_RCVTIMEO,&tv,sizeof(struct timeval));

	sk.sin_family 		= AF_INET;
	sk.sin_addr.s_addr 	= addr;
	sk.sin_port		= HTONS(target_port);
	
	if((databuf = (u8 *) calloc(1,VIDBUFSIZE + 1)) == NULL)
	{
		printf("Cannot allocate %d bytes of memory for video buffer\n",VIDBUFSIZE + 1);
		kill_connection(0);
	}

	clear_reset_screen();
	printf("Making secure connection to %s (Algorithm: %s)\n",inet_ntoa(sk.sin_addr),cameloid_algo_to_name(choice_algo));

	ath_w_p = (cameloid_auth_t *) &init_h.ath;
	
	/*******************************************************
	 ****  Fill in correct authentication information ******
	 *******************************************************/

	cameloid_get_p_random_bytes(urandomfd,nonce,CAMELOID_NONCE_SIZE);
	cameloid_build_auth_packet(ath_w_p,
				   nonce,
				   CAMELOID_PKT_V_INIT);

	init_h.alg 		= choice_algo;
	
	memcpy(init_h.username,username,strlen(username));

	/* obfuscate the username  */
	cameloid_add_gf256(init_h.username,nonce,CAMELOID_MAX_USERNAME);

	if((connect(s,(struct sockaddr *) &sk,sizeof(struct sockaddr_in)))<0)
	{
		printf("Host is not accepting calls at this time\n");
		kill_connection(0);
		return 0;
	}

	if((connect(data_socket,(struct sockaddr *) &sk2, sizeof(struct sockaddr_in)))<0)
	{
		printf("Host is not accepting calls at this time\n");
		kill_connection(0);
		return 0;
	}

	getsockname(data_socket,(struct sockaddr *) &sk3, &sklen);
		
	init_h.data_port = sk3.sin_port;
	
	if((send(s,&init_h,PKT_IN_SIZE,0)) != PKT_IN_SIZE)
	{
		cameloid_print_error("Error sending initial packet.");
		kill_connection(0);
		return 0;
	}
	
	if(verbose)
	{
		printf("Awaiting challenge packet response\n");
		fflush(stdout);
	}

	if((recv(s,buf, PKT_CR_SIZE,0)) != PKT_CR_SIZE)
	{
		printf("Host did not respond in expected time\n");
		kill_connection(0);
		return 0;		
	}
	else
	{
		cr = (cameloid_cr_pkt_t *) buf;
		
		ath_r_p  	= (cameloid_auth_t *) (&cr->ath);
	
		if(cr->ath.pkt_type == CAMELOID_PKT_V_CHALLENGE)
		{
			if(verbose)
			{
				printf("Challenge packet received sending response\n");				
				fflush(stdout);
			}

			cprecv = 1;
			
			/* Verify, decrypt challenge and send back encrypted hash */										
			ptr = cr->data;

			cameloid_hmac_md5(ptr,
					  32,
					  ath_r_p->nonce,
					  CAMELOID_NONCE_SIZE,
					  ctx.keydatahash,
					  CAMELOID_MAC_KEYSIZE,
					  digest);

			if(memcmp(ath_r_p->hmac,digest,16))
			{
				/* This may be too paranoid of an assumption, Hence the word "Possible" */
				#ifdef SUPER_PARANOID
				cameloid_print_error("*Possible* injection attack, Invalid MAC computed for C/R packet."); 
				#endif
				cameloid_print_error("Response not valid, possibly due to incorrect NK specified.");
				close(s);
				return 0;
			}
				
			cameloid_crypt_data(CAMELOID_MODE_DECRYPT,
					    ptr,
					    32,
					    ath_r_p->nonce,&ctx);

			cameloid_generate_hash(ptr,32,hash);
		
			cameloid_get_p_random_bytes(urandomfd,nonce,CAMELOID_NONCE_SIZE);
		
			cameloid_crypt_data(CAMELOID_MODE_ENCRYPT,
					    hash,
					    32,
					    nonce,&ctx);
		
			cameloid_build_cr_packet(&snd_cr_h,hash,nonce,CAMELOID_PKT_V_RESPONSE);
		
			cameloid_hmac_md5(hash,
					  32,
					  nonce,
					  CAMELOID_NONCE_SIZE,
					  ctx.keydatahash,
					  CAMELOID_MAC_KEYSIZE,snd_cr_h.ath.hmac);
		
			cameloid_destroy_memory(hash,32);

			if((send(s,&snd_cr_h,PKT_CR_SIZE,0)) != PKT_CR_SIZE)
			{
				cameloid_print_error("Error sending Challenge/Response Packet.");
				kill_connection(0);
				return 0;
			}

			if((cameloid_net_read(s,buf,PKT_SK_SIZE)) != PKT_SK_SIZE)
			{
				printf("Call disconnected\n"); fflush(stdout);
				close(s);
				return 0;
			}

			
			sk_h = (cameloid_sk_xchg_t *) buf;
			ath_r_p = &sk_h->ath;
			
			if(ath_r_p->pkt_type != CAMELOID_PKT_V_SK)
			{
				printf("Session key segment packet not received\n"); fflush(stdout);
				close(s);	
				return 0;
			}
			else
			{	
				
				/* Verify SK packet, store it and generate our SK */
				if(verbose)
				{
					printf("Session key segment received\n");
					fflush(stdout);
				}
			
				cameloid_hmac_md5(sk_h->sk_segment,
						  32,
						  ath_r_p->nonce,
						  CAMELOID_NONCE_SIZE,
                                          	  ctx.keydatahash,
						  CAMELOID_MAC_KEYSIZE,
						  digest);

	                        if(memcmp(ath_r_p->hmac,digest,16))
        	                {
                	                printf("Possible injection attack, Invalid MAC computed for C/R packet.\n"); fflush(stdout);
                       		        kill_connection(0);
					return 0;
                        	}       
                    
				cameloid_crypt_data(CAMELOID_MODE_DECRYPT,
						    sk_h->sk_segment,
						    CAMELOID_MAX_SK_SEG,
						    ath_r_p->nonce,&ctx);
				
				if(verbose)
				{
					printf("Received SK segment hash -> %32s\n",cameloid_hash_and_show_md5(sk_h->sk_segment,CAMELOID_MAX_SK_SEG));
					fflush(stdout);
				}

				memcpy(SK,sk_h->sk_segment,CAMELOID_MAX_SK_SEG);

				/* Destroy generated session key segment */
				cameloid_destroy_memory(sk_h->sk_segment,CAMELOID_MAX_SK_SEG);				
				
				if(verbose)
				{
					printf("Generating %3d-bit random SK segment\n",CAMELOID_MAX_SK_SEG*8);
				}

				cameloid_get_r_random_bytes(randomfd,SK+CAMELOID_MAX_SK_SEG,CAMELOID_MAX_SK_SEG);
				cameloid_get_p_random_bytes(urandomfd,nonce,CAMELOID_NONCE_SIZE);
				cameloid_build_auth_packet(&sk_w_h.ath,nonce,CAMELOID_PKT_V_SK);
				memcpy(&sk_w_h.sk_segment,SK+CAMELOID_MAX_SK_SEG,CAMELOID_MAX_SK_SEG);
				if(verbose)
				{
					printf("Generated SK segment hash -> %32s\n",
					cameloid_hash_and_show_md5(SK+CAMELOID_MAX_SK_SEG,CAMELOID_MAX_SK_SEG));
				}

				cameloid_crypt_data(CAMELOID_MODE_ENCRYPT,
						    sk_w_h.sk_segment,
						    CAMELOID_MAX_SK_SEG,
						    nonce,&ctx);

				cameloid_hmac_md5(sk_w_h.sk_segment,
						  32,
						  nonce,
						  CAMELOID_NONCE_SIZE,
						  ctx.keydatahash,
                                                  CAMELOID_MAC_KEYSIZE,
						  sk_w_h.ath.hmac);

				cameloid_destroy_memory(nonce,CAMELOID_MAX_SK_SEG);

				/* Destroy old keychain  */
				cameloid_destroy_memory((u8 *) &ctx,sizeof(cameloid_kc_t));
				
				if(verbose)
				{
					printf("Exchanging session key segment\n");
					printf("Session key hash -> %32s\n",cameloid_hash_and_show_md5(SK,CAMELOID_MAX_SK));
					fflush(stdout);
				}
				
				if((send(s,&sk_w_h,PKT_SK_SIZE,0)) != PKT_SK_SIZE)
				{
					cameloid_print_error("Error exchange session key segment.");
					kill_connection(0);
					return 0;
				}
				
				close(s);
				active_socket = data_socket;
				
				/* Run newly created SK through the selected KSA */
				cameloid_key_setup(choice_algo,SK,CAMELOID_MAX_SK,&ctx);
				
				/* Destroy SK buffer */
				cameloid_destroy_memory(SK,CAMELOID_MAX_SK);
				


				/* Show Video Screen */
				printf("Loading X video screen....\n\a");
				daemonize();
				mainwindow = create_mainwindow ();
  				gtk_widget_show (mainwindow);
			
  				imageframe = lookup_widget(mainwindow,"mainscreen");
		
 				while(active_socket)
  				{			
					rdlen = 0;
					if((rdlen = cameloid_net_read(data_socket,databuf,VIDBUFSIZE))<0)
					{
						kill_connection(0);
						return 0;
					}

					if(rdlen < PKT_VDA_SIZE)
					{
						 continue;
						
					}
					
					vdptr = (cameloid_vd_auth_t *) databuf;
					if((strncmp(vdptr->ath.proto_id,CAMELOID_PROTO_IDENT,4)) != 0)
					{		
						continue;
					}
					
					if(vdptr->ath.pkt_type == CAMELOID_PKT_PING)
					{
						memcpy(vdah.ath.proto_id,CAMELOID_PROTO_IDENT,4);
						vdah.ath.pkt_type = CAMELOID_PKT_PONG;
						send(data_socket,&vdah,sizeof(vdah),0);
						gtk_main_iteration_do(FALSE);
						continue;
					}

					if(vdptr->vdlen > (rdlen - (PKT_VDA_SIZE + 8))  || 	
					   vdptr->vdlen < (rdlen - (PKT_VDA_SIZE + 8)) )
					   {
						continue;
					   }
						
					macsize = vdptr->vdlen + (PKT_VDA_SIZE - 20) + 8;
					cameloid_hmac_md5(databuf + 20, 
							  macsize, 
							  NULL, 
							  0,
							  ctx.keydatahash,
							    CAMELOID_MAC_KEYSIZE,hmac);

					if((memcmp(hmac,vdptr->ath.hmac,16)) != 0)
					{
						continue;
					}
						
					memcpy(&ocounter,	databuf + (vdptr->vdlen + PKT_VDA_SIZE) + 0, 4);
					memcpy(&oseq_counter,	databuf + (vdptr->vdlen + PKT_VDA_SIZE) + 4, 4);
 
					counter  		= NTOHL(ocounter);
					seq_counter		= NTOHL(oseq_counter);

					if(counter <= last_counter || seq_counter < last_seqcounter)
					{
						continue;
					}
					
					last_counter 		= counter;
					last_seqcounter		= seq_counter;
					
					
					cameloid_crypt_data(CAMELOID_MODE_DECRYPT,
                                                    databuf+PKT_VDA_SIZE,
                                                    vdptr->vdlen,
                                                    vdptr->ath.nonce,&ctx);
			

					cameloid_decompress_mjpeg(databuf+PKT_VDA_SIZE,rgbptr,vdptr->vdlen);
					pixbuf = gdk_pixbuf_new_from_data(rgbptr,GDK_COLORSPACE_RGB,
									  FALSE,8,320,240,(320 * 3),NULL,NULL);

					gtk_image_set_from_pixbuf((GtkImage *) imageframe,pixbuf);					
        				gtk_main_iteration_do(FALSE);
					gdk_pixbuf_unref(pixbuf);
  				}


			}
				
	
		}
	}		
	
	close(s);	
	cameloid_destroy_memory(buf,RD_BUFFER_SIZE + 1);
	return 0;
	

}

int main(int argc, char **argv)
{
	int c, i;
	u32 addr;
	struct hostent *target_host = NULL;
	u8 *hp = NULL;
	static u8 *username = NULL;

	c = s = addr = use_strong_rand = choice_algo = 0;
	randomfd = urandomfd = verbose = 0;
	send_delay = 1;

	printf("\n\tCAMELOID v%s\n\tSecure Network VideoPhone Client\n\n",VERSION);
	
	if(argc<2)
	{
		printf("Usage: %s -[pduvsc] hostname\n",argv[0]);
		printf("Options:\n");
		printf("\t-p <port>\tAlternate cameloid server port (default port %5d)\n",CAMELOID_VIDEO_LISTEN_PORT);
		printf("\t-d <port>\tAlternate cameloid data port   (default port %5d)\n",CAMELOID_VIDEO_DATA_PORT);  
		printf("\t-u <username>\tUsername of sender\n");
		printf("\t-v\t\tVerbose Mode\n");
		printf("\t-s\t\tUse strong random numbers (rijndael encrypted)\n");
		printf("\t-c <alg>\tSpecify encryption algorithm to use\n\n");
		printf("Supported Algorithms:\n\t");
		for(i = 1; i <= CAMELOID_ALGO_COUNT;i++)
		{
			if(i != CAMELOID_ALGO_COUNT)
			printf("%s, ",cameloid_algo_to_name(i));
			else
			printf("%s ",cameloid_algo_to_name(i));
		}
		
		printf("\n\n");
		return 0;
	}

	while((c = getopt(argc,argv,"u:p:d:c:vs")) != -1)
	{
		switch(c)
		{

			case 'c':
				choice_algo = cameloid_name_to_algo(optarg);
				break;


			case 'u':
				username  = optarg;
				break;

			case 'p':
				target_port = atoi(optarg);
				break;


			case 'd':
				data_port  =  atoi(optarg);
				break;

			case 'v':
				verbose = 1;
				break;

			case 's':
				use_strong_rand = 1;
				break;
		}

	}


	hp = argv[optind];
	if(!hp)
	{
		printf("Please specify a host to call\n\n");
		return 0;
	}

	if((target_port < 0 || target_port > 65535))
	{
		printf("Invalid target port specified\n\n");
		return 0;
	}

	if((data_port  < 0 || data_port  > 65535))
	{
		printf("Invalid data port specified\n\n");
		return 0;
	}

	if(!username)
	{
		printf("Please set your username with the '-u' option\n\n");
		return 0;
	}

	if(strlen(username) > 10)
	{
		printf("Username is too long\n\n");
		return 0;
	}

	if(!(target_host = gethostbyname(hp)))
	{
		cameloid_print_error("Invalid host name specified");
		return 0;
	}


	memcpy(&addr,target_host->h_addr,4);
	
	printf("Initializing\n");

	if(choice_algo<0)
	{
		printf("Invalid encryption algorithm specified.\n\n");
		return 0;
	}

	if(!choice_algo)
		choice_algo = cameloid_get_random_algo();

	if(getuid())
	{
		printf("Please run CAMELOID video client as root (SUID won't work).\n");
		return 0;
	}


	cameloid_lock_all_memory();
	
	/* Lets setup some signals */
	signal(SIGTSTP,	kill_connection);
	signal(SIGQUIT, kill_connection);
        signal(SIGINT,	kill_connection);
        signal(SIGTERM,	kill_connection);
        signal(SIGSEGV,	kill_connection);
        signal(SIGHUP,	kill_connection);
        signal(SIGBUS,	kill_connection);
        signal(SIGFPE,	kill_connection);
        signal(SIGPIPE,	kill_connection);
	signal(SIGCHLD,	SIG_IGN);

	if((cameloid_allocate_memory_struct(&memctx))<0)
		return -1;
	else
	{
		SK = memctx.SK;
		NK = memctx.NK;
	}


	printf("Awaiting Negotiation-Key (NK) to initiate session\n");
	if((cameloid_get_nk(stdin,&memctx))<0)
	{
		printf("Too many failed attempts at entering NK.\n\n");
		return 0;
	}

	printf("Opening random devices %s and %s\n",RANDOM_DEVICE,URANDOM_DEVICE);	
	randomfd 	= cameloid_open_random(); 
	urandomfd 	= cameloid_open_urandom();
	gtk_set_locale ();
        gtk_init (&argc, &argv);
	
 	cameloid_open_conn(addr,username);	
	printf("\n");
	return 0;

}
