/*************************************************
 * CAMELOID Secure End-to-End VideoPhone System
 * Copyright (C) 2005  Dr. Tek
 ************************************************/


#ifdef		VIDEO_CAP


#include "cameloid.h"
#include "cameloid_protocol.h"
#include "server_callbacks.h"

#define		LINE_SIZE			960
#define		FRAME_QUALITY			75
#define		MBUFSIZE			2048

 
typedef struct {
	struct jpeg_destination_mgr pub;
	u8 * outbuffer;
	JOCTET * buffer;
} my_destination_mgr;

typedef my_destination_mgr * my_dest_ptr;


int bytecount;

__inline__
int cameloid_open_video(u8 *videodev)
{
	struct video_channel vc;
	struct video_window  win;

	cameloid_destroy_memory((u8 *) &vc,sizeof(vc));
	cameloid_destroy_memory((u8 *) &win,sizeof(win));
	cameloid_destroy_memory((u8 *) &pic,sizeof(pic));

	if((videofd = open(videodev,O_RDWR))<0)
	{
		printf("Cannot access video device: %s\n",videodev);
		return -1;
	}

	vc.channel = 1;

	if((ioctl(videofd,VIDIOCGCHAN,&vc))<0)
	{
		perror("get video ioctl");
		return -1;
	}
	
	
	vc.channel	= 1;
	vc.type 	= VIDEO_TYPE_CAMERA;
	vc.norm		= VIDEO_MODE_NTSC;
	vc.flags	= 0;

	if((ioctl(videofd,VIDIOCSCHAN,&vc))<0)
	{
		perror("set video ioctl");
		return -1;
	}

	if((ioctl(videofd, VIDIOCGWIN, &win))<0)
	{
		perror("get video ioctl");
		return -1;
	}

	win.x = win.y = win.chromakey = 0;
	win.width 	= 320;
	win.height 	= 240;
	win.flags	= 0;
	
	if((ioctl(videofd, VIDIOCSWIN,  &win))<0)
        {
                perror("set video ioctl");
                return -1;
        }
	
	if((ioctl(videofd, VIDIOCGPICT, &pic))<0)
        {
                perror("get video ioctl");
                return -1;
        }

	ioctl(videofd,VIDIOCGPICT,&pic);

	pic.depth 	= 24;
	pic.palette	= VIDEO_PALETTE_RGB24;
	
	if((ioctl(videofd, VIDIOCSPICT, &pic))<0)
        {
                perror("set video ioctl");
                return -1;
        }
	
	return videofd;
	
}

static void 
decompress_init_source(struct jpeg_decompress_struct *cn)
{
	return;
}

static boolean 
decompress_fill_input_buffer(struct jpeg_decompress_struct  *cn)
{
	return 1;
}

static void 
decompress_skip_input_data(struct jpeg_decompress_struct *cn, long nb)
{
	cn->src->next_input_byte 	+= nb;
	cn->src->bytes_in_buffer 	-= nb;
}

static boolean 
decompress_resync_to_restart(struct jpeg_decompress_struct *cn, int desired)
{
	return 1;
}

static void 
decompress_term_source(struct jpeg_decompress_struct *cn)
{
	return;
}


static void
compress_init_destination(struct jpeg_compress_struct *cn)
{
	my_dest_ptr	dst = (my_dest_ptr) cn->dest;

	dst->buffer 			= malloc(MBUFSIZE);
	dst->pub.next_output_byte	= dst->buffer;
	dst->pub.free_in_buffer		= MBUFSIZE;

}


static void
compress_term_destination(struct jpeg_compress_struct *cn)
{
	my_dest_ptr dst  = (my_dest_ptr) cn->dest;
	int    left = MBUFSIZE - dst->pub.free_in_buffer;

	if(left > 0)
	{
		memcpy(dst->outbuffer + bytecount,dst->buffer,left);
		bytecount += left;
	}	
	
	dst->pub.next_output_byte	= dst->buffer;
	dst->pub.free_in_buffer		= MBUFSIZE;

	return;

}


static boolean
compress_empty_output_buffer(struct jpeg_compress_struct *cn)
{
	my_dest_ptr     dst = (my_dest_ptr) cn->dest;
	
	memcpy(dst->outbuffer + bytecount, dst->buffer, MBUFSIZE);
	bytecount += MBUFSIZE;


	/* !!THIS SHOULD NEVER HAPPEN!! */	
	if(bytecount > COMPRESS_BUF_SIZE)
	{
		bytecount = 0;
	}

	dst->pub.next_output_byte	= dst->buffer;
	dst->pub.free_in_buffer		= MBUFSIZE;

	return TRUE;
}
	
__inline__
int cameloid_decompress_mjpeg(u8 *data, u8 *joutput, u32 size)
{	
	u8 *line;
	struct jpeg_decompress_struct jpeg_cinfo, *cn;
	struct jpeg_error_mgr jpegerr;
	struct jpeg_source_mgr *srcmgr;


	cn = &jpeg_cinfo;

	jpeg_cinfo.err = jpeg_std_error(&jpegerr);
	jpeg_create_decompress(&jpeg_cinfo);

	jpeg_cinfo.src			= (struct jpeg_source_mgr *) (*cn->mem->alloc_small)((j_common_ptr) cn, JPOOL_IMAGE,sizeof(struct jpeg_source_mgr));
	srcmgr				= (struct jpeg_source_mgr *) jpeg_cinfo.src;
	srcmgr->next_input_byte 	= data;
	srcmgr->bytes_in_buffer		= size;
	srcmgr->init_source		= decompress_init_source;
	srcmgr->fill_input_buffer	= decompress_fill_input_buffer;
	srcmgr->skip_input_data		= decompress_skip_input_data;
	srcmgr->resync_to_restart	= decompress_resync_to_restart;
	srcmgr->term_source		= decompress_term_source;


	jpeg_read_header(&jpeg_cinfo,TRUE);

	jpeg_start_decompress(&jpeg_cinfo);	
	while(jpeg_cinfo.output_scanline < jpeg_cinfo.output_height)
	{
		line = joutput + (LINE_SIZE  * jpeg_cinfo.output_scanline);
		jpeg_read_scanlines(&jpeg_cinfo,&line,1);
	}

	jpeg_destroy_decompress(&jpeg_cinfo);
	return 0;		
}

__inline__
int cameloid_compress_mjpeg(u8 *fim, u32 *len, u8 *frameptr)
{
	struct jpeg_compress_struct 	  	jpeg_cnfo, *cn;
	struct jpeg_error_mgr 	 	   	jpegerr;
	my_dest_ptr 	   	   	   	srcmgr;
	u32 compress_frame_size = 0;

	bytecount = 0;
	cn = &jpeg_cnfo;

	jpeg_cnfo.err				=	jpeg_std_error(&jpegerr);
	jpeg_create_compress(cn);
	
	if(!cn->dest)
	{
		cn->dest = (struct jpeg_destination_mgr *)(*cn->mem->alloc_small) ((j_common_ptr) cn, JPOOL_PERMANENT,
									 	  sizeof(my_destination_mgr));
	}

	srcmgr = (my_dest_ptr) 	cn->dest;


	srcmgr->outbuffer                     	= (frameptr + PKT_VDA_SIZE);
        srcmgr->pub.free_in_buffer            	= COMPRESS_BUF_SIZE;
        srcmgr->pub.init_destination          	= compress_init_destination;
        srcmgr->pub.empty_output_buffer       	= compress_empty_output_buffer;
        srcmgr->pub.term_destination          	= compress_term_destination;
	
	cn->image_width  			= 320;
	cn->image_height 			= 240;
	cn->input_components 			= 3;
	cn->in_color_space			= JCS_RGB;

	jpeg_set_defaults(cn);
	jpeg_set_quality(cn, FRAME_QUALITY, TRUE);
	jpeg_start_compress(cn,	TRUE);

	while(cn->next_scanline < cn->image_height)
	{
		jpeg_write_scanlines(cn,(JSAMPARRAY) &fim,1);
		fim += LINE_SIZE;
	}


	jpeg_finish_compress(cn);
	compress_frame_size = bytecount;  
	jpeg_destroy_compress(cn); 
	free(srcmgr->buffer);
	memcpy(len,&compress_frame_size,sizeof(u32));
	return 0;

	
}

__inline__
void swap_rgb24(unsigned char *mem, int n) 
{
        unsigned char  c;
        unsigned char *p = mem;
        register int   i = n ;

        while (--i) 
	{
                c 	= p[0];
                p[0] 	= p[2];
                p[2] 	= c;
                p 	+= 3;
                
        }
             
}

__inline__
int cameloid_read_video(int vidfd, u32 *size, u8 *frameptr, void *map)
{
	static struct video_mmap vidmap;

	vidmap.frame 	= 0;
	vidmap.width 	= 320;
	vidmap.height	= 240;
	vidmap.format	= VIDEO_PALETTE_RGB24;
	
	if((ioctl(vidfd,VIDIOCMCAPTURE,&vidmap))<0)
	{
		perror("ioctl");
		return -1;
	}
	if((ioctl(vidfd,VIDIOCSYNC,&vidmap))<0)
	{
		perror("ioctl");
		return -1;
	}

	swap_rgb24(map,320 * 240);
	return cameloid_compress_mjpeg(map,size,frameptr); 
	
}

#endif
