/*************************************************
 * CAMELOID Secure End-to-End VideoPhone System
 * Copyright (C) 2005  Dr. Tek                     
 ************************************************/

#ifndef		VIDEO_CAP
#define		VIDEO_CAP
#endif

#ifndef         LINUX
#error          CAMELOID video currently works on linux only.. Sorry :(
#endif

#include "cameloid.h"
#include "cameloid_memory.h"
#include "cameloid_voice.h"
#include "cameloid_protocol.h"
#include "cameloid_functions.h"
#include <gtk/gtk.h>
#include "server_interface.h"
#include "server_callbacks.h"
#include "support.h"

int 	randomfd, urandomfd, server_socket,
	vidfd,    serv_port, data_port, pause_state;


cameloid_kc_t 		*keystruct = NULL;
cameloid_mstruct_t	 memctx;
u8 			*video_dev  		= CAMELOID_VIDEO_DEVICE;


__inline__
void kill_destroy_buffers(void)
{

	cameloid_destroy_memory(SK,CAMELOID_MAX_SK);
        cameloid_destroy_memory(NK,CAMELOID_MAX_NK);	

	if(keystruct)
                cameloid_destroy_memory((u8 *) keystruct, sizeof(cameloid_kc_t));

}


__inline__
void enable_video_pause(int sig)
{
		
	if(active_socket)
	{
		if(!pause_state)
		{
			pause_state = 1;
			do_system_beep();
			fflush(stdout);
			
		}
		else
		{
			pause_state = 0;
		}
	}
	
}

__inline__
void kill_destroy_sockets(void)
{
	if(active_socket)
	{
		close(active_socket);
		active_socket = 0;
	}


}


__inline__
void kill_connection(int sig)
{
	struct termios termset;

	kill_destroy_buffers();

	/* This is done to ensure that the terminal will
           have the echo flags restored in case of a pre-mature
           cancellation */ 

         tcgetattr(fileno(stdin),&termset);
         termset.c_lflag |= (ECHO | ECHOE | ECHOK | ECHONL);
         tcsetattr(fileno(stdin),TCSANOW,&termset);
	
	if(sig != SIGSEGV)
	{
		if(active_socket || sig == SIGPIPE)
		{
			if(active_socket) 
			{
				printf("\nConnection Terminated\n"); fflush(stdout);
				kill_destroy_sockets();
				active_socket = 0;
			}
			return;
		}
	}
        else
	{
		printf("Segmentation Fault Captured\n"); fflush(stdout);
	}

	kill_destroy_sockets();


	if(!active_socket && (sig == SIGINT || sig == SIGTERM || sig == SIGHUP || sig == SIGSEGV))
	{
		cameloid_destroy_memory_struct(&memctx);
		printf("Shutting down server\n\n");
		close(server_socket);
		close(randomfd);
		close(urandomfd);
		exit(0);
	}
}


__inline__
void cameloid_video_wait(int s, int data_socket)
{
	int 		 	 sklen = sizeof(struct sockaddr_in), 
				 use_algo, pt, size, r_skt = 0, ret = 0;

	u8 			 *buf, *nonce, *digest, 
				 username[CAMELOID_MAX_USERNAME + 1], *frameptr = NULL, rsp;

	u16 			 sender_data_port;
	struct sockaddr_in 	 sk;
	cameloid_auth_t 	*auth_h;
	cameloid_vd_auth_t	*vdptr,vdah;

	cameloid_init_pkt_t 	*init_h;
	cameloid_cr_pkt_t 	*cr_h, snd_cr_h;
	cameloid_challenge_t 	 chal;
	cameloid_sk_xchg_t	 sk_h;
	cameloid_kc_t 		 ctx;
	struct timeval	 	 tv;
	u32			 vdatalen, counter, seq_counter, ocounter, oseq_counter,macsize;
	GtkWidget		*mainwindow, *imageframe,*alertwin;
	GdkPixbuf		*pixbuf;
	u8			*rgbptr;
	
	tv.tv_sec 		= CAMELOID_NET_TIMEOUT;
	tv.tv_usec		= 0;


	buf	= (u8 *) cameloid_alloc(RD_BUFFER_SIZE + 1);
	if(!buf)
	{
		return;
	}

	nonce	= (u8 *) cameloid_alloc(CAMELOID_NONCE_SIZE + 1);
	if(!nonce)
	{
		return;
	}

	digest	= (u8 *) cameloid_alloc(16); 
	if(!digest)
	{
		return;
	}
	
	if((frameptr = (u8 *) cameloid_alloc(COMPRESS_BUF_SIZE + PKT_VDA_SIZE + 8 + 1)) == NULL)
         {
                printf("Cannot allocate %d bytes for video stream\n",COMPRESS_BUF_SIZE + PKT_VDA_SIZE + 8 + 1);
                kill_connection(0);
		return;
         }

	if((rgbptr = mmap(0, (320 * 240)*3, PROT_READ|PROT_WRITE, MAP_SHARED, vidfd, 0)) == NULL)
	{
		printf("Cannot allocate %d bytes of memory for rgb buffer\n",(320*240)*3);
		kill_connection(0);
	}

	mainwindow   = create_mainwindow();
	alertwin     = create_alertwin();
	imageframe   = lookup_widget(mainwindow,"mainscreen");

	keystruct = &ctx;
	while(1)
	{
		
		gtk_widget_hide_all(mainwindow);
		
		while(gtk_events_pending())
		{
			gtk_main_iteration();
		}

		if(r_skt)
		{
			close(r_skt);
		}

		pause_state = vdatalen    = ocounter = oseq_counter = counter 	= 0;
		seq_counter = sender_data_port 	= use_algo =  size = pt = rsp =  0;
		r_skt   = active_socket 	= 0;

		clear_reset_screen(); 
		cameloid_destroy_memory(digest,16);
		cameloid_destroy_memory(nonce,CAMELOID_NONCE_SIZE + 1);
		cameloid_destroy_memory(frameptr,COMPRESS_BUF_SIZE + PKT_VDA_SIZE + 8 + 1);
		cameloid_destroy_memory(username,CAMELOID_MAX_USERNAME+1);
		cameloid_destroy_memory(NK,CAMELOID_MAX_NK);
		cameloid_destroy_memory(SK,CAMELOID_MAX_SK);
		cameloid_destroy_memory((u8 *)&sk,sizeof(struct sockaddr_in));
				
		cameloid_destroy_memory(buf,RD_BUFFER_SIZE);
		cameloid_destroy_memory((u8 *) keystruct, sizeof(cameloid_kc_t));
		printf("Awaiting secure video streaming request\n");
	
		if((listen(s,1))<0)
		{
			continue;
		}
	
		if((r_skt = accept(s,(struct sockaddr *) &sk,&sklen))<0)
		{
			continue;
		}

		setsockopt(r_skt,SOL_SOCKET,SO_RCVTIMEO,&tv,sizeof(struct timeval));

		if((size = cameloid_net_read(r_skt,buf,PKT_IN_SIZE)) != PKT_IN_SIZE)
		{
			close(r_skt);
			continue;
		}
					
		auth_h = (cameloid_auth_t *) buf;
		pt   = auth_h->pkt_type;
	
		if(pt != CAMELOID_PKT_V_INIT)
		{
			close(r_skt);
			continue;
		}

		if(memcmp(auth_h->proto_id,CAMELOID_PROTO_IDENT,4))
		{
			close(r_skt);
			continue;
		}	
		else
		{   	
				
			init_h   = (cameloid_init_pkt_t *) (buf);
			use_algo = init_h->alg;

 			if(!cameloid_algo_to_name(use_algo))
			{
				close(r_skt);
				continue;
			}

			cameloid_destroy_memory(username,CAMELOID_MAX_USERNAME);
			memcpy(username,init_h->username,CAMELOID_MAX_USERNAME);

			/* de-obfuscate the username */
		        cameloid_add_gf256(username,init_h->ath.nonce,CAMELOID_MAX_USERNAME);

			gtk_widget_show(alertwin);
                        while(!okay_button_clicked)
                        {
                                gtk_main_iteration();
                                continue;
                        }
                        okay_button_clicked = 0;

                        while(gtk_events_pending())
                         {
                              gtk_main_iteration();
                         }

			printf("User %s from %s wants to establish a secure video connection\nThe algorithm chosen was: %s\n",
				username,inet_ntoa(sk.sin_addr),cameloid_algo_to_name(use_algo));
			do_system_beep();

			printf("Do you want to accept the call [Y]es [N]o ? "); fflush(stdout);
						
			while(rsp != 'Y' && rsp != 'N')
			{
				rsp = toupper(getchar());
			}
									
			if(rsp == 'N')
			{
				printf("Cancelling request by user %s\n",init_h->username);
				close(r_skt);
				continue;
			}

			if(rsp == 'Y')
			{
				active_socket 	 = r_skt;
				use_algo 	 = init_h->alg;
				sender_data_port = init_h->data_port;
			
				getchar();
	                        if((cameloid_get_nk(stdin,&memctx))<0)
				{
					printf("Too many failed NK attempts.\n");
					fflush(stdout);
					sleep(1);
					kill_connection(0);
					continue;
				}
								
        	                cameloid_key_setup(use_algo,NK,strlen(NK),&ctx);
				cameloid_destroy_memory(NK,strlen(NK));
				printf("Initiating new secure session with %s\n",inet_ntoa(sk.sin_addr)); 
				printf("Generating random challenge\n");

				/* Generate our random challenge */			
				cameloid_get_p_random_bytes(randomfd,chal.chal,32);
				cameloid_get_p_random_bytes(urandomfd,nonce,CAMELOID_NONCE_SIZE);
			        cameloid_generate_hash(chal.chal,32,chal.hash);
   
        			cameloid_crypt_data(CAMELOID_MODE_ENCRYPT,
						    chal.chal,
						    32,
						    nonce,&ctx);	
				
				printf("Sending random challenge and awaiting response\n"); 
				cameloid_build_cr_packet(&snd_cr_h,chal.chal,nonce,CAMELOID_PKT_V_CHALLENGE);

				cameloid_hmac_md5(snd_cr_h.data,32,nonce,CAMELOID_NONCE_SIZE,
						  ctx.keydatahash,CAMELOID_MAC_KEYSIZE,snd_cr_h.ath.hmac);

				if((send(r_skt,&snd_cr_h,PKT_CR_SIZE,0))<0)
				{
					kill_connection(0);
					continue;
				}

				if((cameloid_net_read(r_skt,buf,PKT_CR_SIZE)) != PKT_CR_SIZE)
				{
					kill_connection(0);
					continue;
				}
				
				cr_h = (cameloid_cr_pkt_t *) buf;				
				auth_h = &cr_h->ath;
				
				if(auth_h->pkt_type != CAMELOID_PKT_V_RESPONSE)
				{
					cameloid_print_error("Unexpected packet received");
					kill_connection(0);
					continue;
				}
				else
				{
					cameloid_hmac_md5(cr_h->data,
							  32,
							  auth_h->nonce,
							  CAMELOID_NONCE_SIZE,
							  ctx.keydatahash,
							  CAMELOID_MAC_KEYSIZE,
							  digest);
			
					if(memcmp(digest,auth_h->hmac,16))
					{
						 cameloid_print_error("*Possible* injection attack, Invalid MAC computed for C/R packet.");
						 kill_connection(0);
						 continue;
					} 

					cameloid_crypt_data(CAMELOID_MODE_DECRYPT,
							    cr_h->data,
							    32,
							    auth_h->nonce,&ctx);
				
					if(memcmp(cr_h->data,chal.hash,32))
					{
						printf("Invalid response to challenge\n");
						kill_connection(0);
						continue;
					}
					else
					{
						printf("Response to challenge is valid\n");
						printf("Generating %3d-bit random SK\n",CAMELOID_MAX_SK_SEG*8);
						cameloid_get_r_random_bytes(randomfd,SK,CAMELOID_MAX_SK_SEG);

						printf("Generated SK segment hash -> %32s\n",cameloid_hash_and_show_md5(SK,CAMELOID_MAX_SK_SEG));
						cameloid_get_p_random_bytes(urandomfd,nonce,CAMELOID_NONCE_SIZE);

						memcpy(sk_h.sk_segment,SK,CAMELOID_MAX_SK_SEG);
						cameloid_crypt_data(CAMELOID_MODE_ENCRYPT,
								    sk_h.sk_segment,
								    CAMELOID_MAX_SK_SEG,
								    nonce,&ctx);
					
						cameloid_build_auth_packet(&sk_h.ath,nonce,CAMELOID_PKT_V_SK);																
						
						cameloid_hmac_md5(sk_h.sk_segment,
								  32, 
								  nonce, 
								  CAMELOID_NONCE_SIZE,
								  ctx.keydatahash,
                                                  		  CAMELOID_MAC_KEYSIZE,
								  sk_h.ath.hmac);

						if((send(r_skt,&sk_h,PKT_SK_SIZE,0)) != PKT_SK_SIZE)
						{
							kill_connection(0);
							continue;
						}
					
						printf("Awaiting SK segment from %s\n",inet_ntoa(sk.sin_addr));
					
						if((cameloid_net_read(r_skt,(u8 *) &sk_h,PKT_SK_SIZE)) != PKT_SK_SIZE)
						{
							kill_connection(0);
							continue;
						}
						
						if(sk_h.ath.pkt_type == CAMELOID_PKT_V_SK)
						{
							 cameloid_hmac_md5(sk_h.sk_segment,
									   32,
									   sk_h.ath.nonce,
									   CAMELOID_NONCE_SIZE,
                                          				   ctx.keydatahash,
									   CAMELOID_MAC_KEYSIZE,
									   digest);
                        
                              		  		if(memcmp(sk_h.ath.hmac,digest,16))
                                			{
                                	       	 		printf("Possible injection attack, Invalid MAC computed for C/R packet.\n"); fflush(stdout);
                        	                		kill_connection(0);
                	                        		continue;
		                                	}

							printf("Session key segment received\n");
							cameloid_crypt_data(CAMELOID_MODE_DECRYPT,
									    sk_h.sk_segment,
									    CAMELOID_MAX_SK_SEG,
									    sk_h.ath.nonce,&ctx);

							/* Destroy old keychain */
							cameloid_destroy_memory((u8 *) &ctx,sizeof(cameloid_kc_t));
						
							printf("Received SK segment hash -> %32s\n",cameloid_hash_and_show_md5(sk_h.sk_segment,CAMELOID_MAX_SK_SEG));
							memcpy(SK+CAMELOID_MAX_SK_SEG,sk_h.sk_segment,CAMELOID_MAX_SK);
							printf("Session key hash -> %32s\n",cameloid_hash_and_show_md5(SK,CAMELOID_MAX_SK));


							/* Run newly created SK through the selected KSA */
							cameloid_key_setup(use_algo,SK,CAMELOID_MAX_SK,&ctx);

							/* Destroy SK buffer */
							cameloid_destroy_memory(SK,CAMELOID_MAX_SK);
							close(r_skt);
							active_socket = 255;

							sk.sin_port			= sender_data_port;
								
							/* Set socket into a connected state */
							connect(data_socket,&sk,sizeof(sk));

							printf("Securely transmitting video stream to user %s\n",username); fflush(stdout);

							gtk_widget_show_all(mainwindow);
							
							 while(gtk_events_pending())
					                 {
                        					gtk_main_iteration();
                					 }

							while(active_socket)
							{	
								
								/* Send a ping to our caller every couple of packets to ensure */
								/* that he is still listening */

								vdatalen = 0;
								gtk_main_iteration_do(FALSE);
								if(pause_state)
								{
									cameloid_destroy_memory(buf,sizeof(buf));
									cameloid_destroy_memory((u8 *) &vdah,sizeof(vdah));
									vdptr 		    = (cameloid_vd_auth_t *) buf;
	
									vdptr->ath.pkt_type = CAMELOID_PKT_PING;
									memcpy(vdptr->ath.proto_id,CAMELOID_PROTO_IDENT,4);
									
									/* If we are paused then keep sending a bunch of ping's */
									while(pause_state && active_socket)
									{
										
										if((send(data_socket,buf,PKT_VDA_SIZE,0))<0)
										{
											kill_connection(0);
											break;
										}
											

										gtk_main_iteration_do(FALSE);
										usleep(50000);
										continue;
									}

									
								}
								

								if(!pause_state)
								{
									ret = cameloid_read_video(vidfd, &vdatalen, frameptr,rgbptr);
								
									if(ret<0)
									{
										continue;
									}

									pixbuf = gdk_pixbuf_new_from_data(rgbptr,GDK_COLORSPACE_RGB,
                                                                          				  FALSE,8,320,240,(320 * 3),NULL,NULL);	
						
									gtk_image_set_from_pixbuf((GtkImage *) imageframe,pixbuf);  
									gtk_main_iteration_do(FALSE);
                                        				gdk_pixbuf_unref(pixbuf);
									
								}								

								if(!ret)	
								{	
									vdptr = (cameloid_vd_auth_t *) frameptr;	
									vdptr->vdlen = vdatalen;
									memcpy(vdptr->ath.proto_id,CAMELOID_PROTO_IDENT,4);
									
									cameloid_get_p_random_bytes(urandomfd,vdptr->ath.nonce,CAMELOID_NONCE_SIZE);
									cameloid_crypt_data(CAMELOID_MODE_ENCRYPT,
                                                                            		    frameptr + PKT_VDA_SIZE,
                                                                            		    vdatalen,
                                                                            		    vdptr->ath.nonce,&ctx);

									counter++;		
									ocounter = HTONL(counter);	
									if(!counter)
									{
										seq_counter++;
										oseq_counter = HTONL(seq_counter);
									}

									memcpy(frameptr + vdatalen + PKT_VDA_SIZE + 0,&ocounter,4);
									memcpy(frameptr + vdatalen + PKT_VDA_SIZE + 4,&oseq_counter,4);
										
									macsize = (vdatalen + (PKT_VDA_SIZE - 20) + 8);

									/* Do our HMAC-MD5 calculations frameptr + 20 to ensure we */ 
									/* hit everything past our hmac fields 		      */
									cameloid_hmac_md5(frameptr + 20, macsize, NULL,0,ctx.keydatahash,
											  CAMELOID_MAC_KEYSIZE,vdptr->ath.hmac);
												
									if((send(data_socket,frameptr,vdatalen + PKT_VDA_SIZE + 8,0))<0)
									{
										kill_connection(0);
										break;
									}
								}					
                                                 	         
                                       				
							}
						}
						
					}
				
				}		
						
								
			}
												
		}
					
	}
											
}	


__inline__
void cameloid_bind_port(void)
{
	int s, data_socket, out_socket, v = 1, bufsize = 40000;
	struct sockaddr_in sk;
	struct timeval tv;


	s = data_socket = out_socket = 0;

	if((s 			= socket(AF_INET,SOCK_STREAM,0))<0)
	{
		cameloid_print_error("Cannot open TCP network socket.");
		return;
	}

	if((data_socket 	= socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP))<0)
	{
		cameloid_print_error("Cannot open UDP network socket.");
		return;
	}

	sk.sin_port 		= HTONS(serv_port);
	sk.sin_addr.s_addr 	= 0;
	sk.sin_family 		= AF_INET;
	
	setsockopt(s, SOL_SOCKET,SO_REUSEADDR, (char*)&v, sizeof(v));
	if((bind(s,(struct sockaddr *) &sk, sizeof(struct sockaddr_in)))<0)
	{
		printf("Cannot bind to TCP port %5d\n",serv_port);
		return;
	}

	sk.sin_port		= HTONS(data_port);
	setsockopt(data_socket, SOL_SOCKET,SO_REUSEADDR, (char*)&v, sizeof(v));
	setsockopt(out_socket,  SOL_SOCKET,SO_SNDBUF, &bufsize, sizeof(int));
	setsockopt(data_socket, SOL_SOCKET,SO_RCVBUF, &bufsize, sizeof(int));
	
	if((bind(data_socket,(struct sockaddr *) &sk, sizeof(struct sockaddr_in)))<0)
	{
		printf("Cannot bind to UDP port %5d\n",data_port);
		return;
	}

	tv.tv_sec 	= CAMELOID_NET_UDP_TIMEOUT;
	tv.tv_usec	= 0;
	setsockopt(data_socket,SOL_SOCKET,SO_RCVTIMEO,&tv,sizeof(struct timeval));
	server_socket = 0;
	cameloid_video_wait(s,data_socket);	
	
}

int main(int argc, char **argv)
{
	int j = 0;
	
	use_strong_rand = randomfd = urandomfd = server_socket = serv_port = data_port = 0;
	printf("\n\tCAMELOID v%s\n\tSecure Network VideoPhone Server\n\n",VERSION);

	send_delay = 1;
	close(2);

	if(argc > 1)
	{
		while(( j =getopt(argc,argv,"p:d:s:v:h")) != -1)
		{

			switch(j)
			{

				case 'h':
		
					printf("usage: %s [options]\n\nOptions:\n",argv[0]);
					printf("\t-p[port]	Bind to an alternate server port (default: %5d)\n",CAMELOID_VIDEO_LISTEN_PORT); 
					printf("\t-d[port]	Bind to an alternate data port   (default: %5d)\n",CAMELOID_VIDEO_DATA_PORT);
					printf("\t-s		Use strong random numbers (rijndael encrypted)\n");
					printf("\t-v[device]	Specify a video device (default is: %s)\n",CAMELOID_VIDEO_DEVICE);
					printf("\t-h		Show this help menu\n\n");
					return 0;
					break;

				case 'j':
			
					use_strong_rand = 1;
					break;

			
		
				case 'v':
				
					video_dev	= optarg;
					break;
				
				case 'p':
		

					serv_port = atoi(optarg);
					if(serv_port<0 || serv_port > 65535)
					{
						printf("Invalid port specified\n");
						return 0;
					}
					break;


				
				case 'd':
				
					serv_port = atoi(optarg);
                              	 	if(serv_port<0 || serv_port > 65535)
                                	{
                                        	printf("Invalid port specified\n");
                                        	return 0;
                                	}
					break;

			}

		}

	}
	

	if(getuid())
        {
                printf("Please run CAMELOID video server as root (SUID won't work).\n");
                return 0;
        }


	printf("Initializing\n");

	cameloid_lock_all_memory();	
        
	if((vidfd = cameloid_open_video(video_dev))<0)
	{
		return 0;	
	}
	
	if((cameloid_allocate_memory_struct(&memctx))<0)
		return -1;
	else
	{
		NK 	= memctx.NK;
		SK 	= memctx.SK;
	}
	printf("Opening random device %s\n",RANDOM_DEVICE);
	
	randomfd  	= cameloid_open_random(); 
	urandomfd	= cameloid_open_urandom();


	if(!serv_port)	serv_port	=	CAMELOID_VIDEO_LISTEN_PORT;
	if(!data_port)	data_port	= 	CAMELOID_VIDEO_DATA_PORT;
	
	signal(SIGINT,	kill_connection);
	signal(SIGSEGV,	kill_connection);  
	signal(SIGTERM,	kill_connection);
	signal(SIGHUP,	kill_connection);
	signal(SIGFPE,	kill_connection);
	signal(SIGBUS,	kill_connection);
	signal(SIGTSTP,	enable_video_pause);	
	signal(SIGPIPE,	kill_connection);
	signal(SIGQUIT,	kill_connection);    

	gtk_set_locale();
	gtk_init(&argc,&argv);	
	cameloid_bind_port();
	return 0;

}
